import unittest
import tempfile
import shutil
from pathlib import Path

from mockito import when, any, verify, unstub, mock

from bcipy.helpers.visualization import visualize_session_data
from bcipy.helpers import visualization
from bcipy.helpers.raw_data import RawData
from bcipy.helpers.load import load_json_parameters
from bcipy.config import DEFAULT_PARAMETERS_PATH


class TestVisualizeSessionData(unittest.TestCase):
    """Test Session Data Visualization."""

    def setUp(self):
        self.tmp_dir = Path(tempfile.mkdtemp())
        self.parameters = load_json_parameters(DEFAULT_PARAMETERS_PATH, value_cast=True)
        self.raw_data_mock = mock()
        self.raw_data_mock.daq_type = 'DSI-24'
        self.raw_data_mock.sample_rate = 300
        self.channel_map_mock = mock()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        unstub()

    def test_visualize_session_data(self):
        """Test visualize_session_data with default parameters."""
        trigger_timing_mock = [1, 2]
        trigger_label_mock = ['target', 'nontarget']
        show = True
        when(RawData).load(any()).thenReturn(self.raw_data_mock)
        when(visualization).analysis_channels(any(), any()).thenReturn(self.channel_map_mock)
        when(visualization).trigger_decoder(
            offset=any(), trigger_path=any(), exclusion=any()).thenReturn(
            (trigger_label_mock, trigger_timing_mock, []))
        when(visualization).visualize_erp(
            self.raw_data_mock,
            self.channel_map_mock,
            trigger_timing_mock,
            any(),
            any(),
            transform=any(),
            plot_average=any(),
            plot_topomaps=any(),
            save_path=any(),
            show=show).thenReturn(True)

        # This fuction returns a figure handle directly from visualization.visualize_erp,
        # however, we are mocking it to return True here
        resp = visualize_session_data(self.tmp_dir, self.parameters, show=show)
        self.assertTrue(resp)
        verify(RawData, times=1).load(any())
        verify(visualization, times=1).analysis_channels(any(), any())
        verify(visualization, times=1).trigger_decoder(offset=any(), trigger_path=any(), exclusion=any())
        verify(visualization, times=1).visualize_erp(
            self.raw_data_mock,
            self.channel_map_mock,
            trigger_timing_mock,
            any(),
            any(),
            transform=any(),
            plot_average=any(),
            plot_topomaps=any(),
            save_path=any(),
            show=show)

    def test_visualize_session_data_with_no_valid_targets(self):
        """Test visualize_session_data throws an error if there are no valid targets."""
        trigger_timing_mock = [1, 2]
        # don't include target in the trigger labels
        trigger_label_mock = ['nontarget', 'nontarget']
        show = False
        when(RawData).load(any()).thenReturn(self.raw_data_mock)
        when(visualization).analysis_channels(any(), any()).thenReturn(self.channel_map_mock)
        when(visualization).trigger_decoder(
            offset=any(), trigger_path=any(), exclusion=any()).thenReturn(
            (trigger_label_mock, trigger_timing_mock, []))

        with self.assertRaises(AssertionError):
            visualize_session_data(self.tmp_dir, self.parameters, show=show)

    def test_visualize_session_data_with_no_valid_nontargets(self):
        """Test visualize_session_data throws an error if there are no valid nontargets."""
        trigger_timing_mock = [1, 2]
        # don't include nontarget in the trigger labels
        trigger_label_mock = ['target', 'target']
        show = False
        when(RawData).load(any()).thenReturn(self.raw_data_mock)
        when(visualization).analysis_channels(any(), any()).thenReturn(self.channel_map_mock)
        when(visualization).trigger_decoder(
            offset=any(), trigger_path=any(), exclusion=any()).thenReturn(
            (trigger_label_mock, trigger_timing_mock, []))

        with self.assertRaises(AssertionError):
            visualize_session_data(self.tmp_dir, self.parameters, show=show)

    def test_visualize_session_data_with_invalid_timing(self):
        """Test visualize_session_data throws an error if there are invalid trigger times."""
        # in this case, leave out the second trigger time
        trigger_timing_mock = [1]
        trigger_label_mock = ['target', 'nontarget']
        show = False
        when(RawData).load(any()).thenReturn(self.raw_data_mock)
        when(visualization).analysis_channels(any(), any()).thenReturn(self.channel_map_mock)
        when(visualization).trigger_decoder(
            offset=any(), trigger_path=any(), exclusion=any()).thenReturn(
            (trigger_label_mock, trigger_timing_mock, []))

        with self.assertRaises(AssertionError):
            visualize_session_data(self.tmp_dir, self.parameters, show=show)

        # in this case, add a third trigger time
        trigger_timing_mock = [1, 2, 3]
        when(visualization).trigger_decoder(
            offset=any(), trigger_path=any(), exclusion=any()).thenReturn(
            (trigger_label_mock, trigger_timing_mock, []))

        with self.assertRaises(AssertionError):
            visualize_session_data(self.tmp_dir, self.parameters, show=show)


if __name__ == "__main__":
    unittest.main()



import errno
import json
import os
from pathlib import Path
from shutil import copyfile
from time import localtime, strftime
from typing import Any, List

from bcipy.acquisition.devices import DeviceSpec
from bcipy.config import (DEFAULT_ENCODING, DEFAULT_EXPERIMENT_ID,
                          DEFAULT_LM_PARAMETERS_FILENAME,
                          DEFAULT_LM_PARAMETERS_PATH,
                          DEFAULT_PARAMETER_FILENAME)
from bcipy.helpers.validate import validate_experiments


def save_json_data(data: Any, location: str, name: str) -> str:
    """
    Writes Parameters as a json file.

        data: any data that can be dumped as json
        location[str]: directory in which to save
        name[str]: optional name of file; default is parameters.json

    Returns path of saved file
    """
    path = Path(location, name)
    with open(Path(location, name), 'w', encoding=DEFAULT_ENCODING) as json_file:
        json.dump(data, json_file, ensure_ascii=False, indent=2)
    return str(path)


def save_experiment_data(experiments, fields, location, name) -> str:
    validate_experiments(experiments, fields)
    return save_json_data(experiments, location, name)


def save_field_data(fields, location, name) -> str:
    return save_json_data(fields, location, name)


def save_experiment_field_data(data, location, name) -> str:
    return save_json_data(data, location, name)


def init_save_data_structure(data_save_path: str,
                             user_id: str,
                             parameters: str,
                             task: str,
                             experiment_id: str = DEFAULT_EXPERIMENT_ID
                             ) -> str:
    """
    Initialize Save Data Structure.

        data_save_path[str]: string of path to save our data in
        user_id[str]: string of user name / related information
        parameters[str]: parameter location for the experiment
        experiment_id[str]: Name of the experiment. Default name is DEFAULT_EXPERIMENT_ID.
        task[str]: name of task type. Ex. RSVP Calibration

    """

    # make an experiment folder : note datetime is in utc
    save_folder_name = f'{data_save_path}{experiment_id}/{user_id}'
    dt = strftime('%a_%d_%b_%Y_%Hhr%Mmin%Ssec_%z', localtime())
    task = task.replace(' ', '_')
    save_directory = f"{save_folder_name}/{user_id}_{task}_{dt}"

    try:
        # make a directory to save data to
        os.makedirs(save_folder_name)
        os.makedirs(save_directory)
        os.makedirs(os.path.join(save_directory, 'logs'), exist_ok=True)

    except OSError as error:
        # If the error is anything other than file existing, raise an error
        if error.errno != errno.EEXIST:
            raise error

        # since this is only called on init, we can make another folder run
        os.makedirs(save_directory)
        os.makedirs(os.path.join(save_directory, 'logs'), exist_ok=True)

    copyfile(parameters, Path(save_directory, DEFAULT_PARAMETER_FILENAME))

    copyfile(DEFAULT_LM_PARAMETERS_PATH, Path(save_directory, DEFAULT_LM_PARAMETERS_FILENAME))

    return save_directory


def save_device_specs(device_specs: List[DeviceSpec], location: str,
                      name: str) -> str:
    """
    Save device spec to a json file.

    Parameters:
        device_spec (DeviceSpec): device spec to save
        location (str): location to save the file
        name (str): name of the file to be saved

    Returns:
        str: path to the saved file
    """
    return save_json_data([spec.to_dict() for spec in device_specs], location,
                          name)


def _save_session_related_data(file, session_dictionary):
    """
    Save Session Related Data.
    Parameters
    ----------
        file[str]: string of path to save our data in
        session_dictionary[dict]: dictionary of session data. It will appear
            as follows:
                {{ "series": {
                        "1": {
                          "0": {
                            "target_text": "COPY_PHRASE",
                            "current_text": "COPY_",
                            "eeg_len": 22,
                            "next_display_state": "COPY_",
                            "stimuli": [["+", "_", "G", "L", "B"]],
                            "target_info": [
                              "nontarget", ... ,
                            ],
                            "timing_sti": [[1, 0.2, 0.2, 0.2, 0.2]],
                            "triggers": [[ "+", 0.0], ["_", 0.9922] ..],
                            },
                        ... ,
                        "7": {
                            ... ,
                  },
                  "mode": "RSVP",
                  "session": "data/demo_user/demo_user",
                  "task": "Copy Phrase",
                  "total_time_spent": 83.24798703193665
                }}
    Returns
    -------
        file, session data file (json file)
    """
    # Try opening as json, if not able to use open() to create first
    try:
        file = json.load(file, 'wt')
    except BaseException:
        file = open(file, 'wt', encoding=DEFAULT_ENCODING)

    # Use the file to dump data to
    json.dump(session_dictionary, file, indent=2)
    return file



"""Helper functions for managing and parsing session.json data."""
import csv
import itertools
import json
import os
import sqlite3
import subprocess
from dataclasses import dataclass, fields
from typing import Any, Dict, List

import openpyxl
from openpyxl.chart import BarChart, Reference
from openpyxl.styles import PatternFill
from openpyxl.styles.borders import BORDER_THIN, Border, Side
from openpyxl.styles.colors import BLACK, WHITE, YELLOW

from bcipy.config import (
    BCIPY_ROOT,
    DEFAULT_ENCODING,
    DEFAULT_PARAMETER_FILENAME,
    EXPERIMENT_DATA_FILENAME,
    SESSION_DATA_FILENAME,
    SESSION_SUMMARY_FILENAME,
)
from bcipy.helpers.load import (load_experiment_fields, load_experiments,
                                load_json_parameters)
from bcipy.helpers.validate import validate_field_data_written
from bcipy.task.data import Session


def read_session(file_name=SESSION_DATA_FILENAME) -> Session:
    """Read the session data from the given file."""
    with open(file_name, 'r', encoding=DEFAULT_ENCODING) as json_file:
        return Session.from_dict(json.load(json_file))


def session_data(data_dir: str) -> Dict:
    """Returns a dict of session data transformed to map the alphabet letter
    to the likelihood when presenting the evidence. Also removes attributes
    not useful for debugging."""

    parameters = load_json_parameters(os.path.join(data_dir,
                                                   DEFAULT_PARAMETER_FILENAME),
                                      value_cast=True)
    session = read_session(os.path.join(data_dir, SESSION_DATA_FILENAME))
    data = session.as_dict(evidence_only=True)
    data['target_text'] = parameters['task_text']
    return data


def collect_experiment_field_data(experiment_name,
                                  save_path,
                                  file_name=EXPERIMENT_DATA_FILENAME) -> None:
    experiment = load_experiments()[experiment_name]
    experiment_fields = load_experiment_fields(experiment)

    if experiment_fields:
        cmd = (f'python {BCIPY_ROOT}/gui/experiments/ExperimentField.py -e'
               f' "{experiment_name}" -p "{save_path}" -f "{file_name}"')
        subprocess.check_call(cmd, shell=True)
        # verify data was written before moving on
        if not validate_field_data_written(save_path, file_name):
            raise Exception('Field data not collected!')


@dataclass(frozen=True)
class EvidenceRecord:
    """Record summarizing Inquiry evidence."""
    series: int
    inquiry: int
    stim: str
    lm: float
    eeg: float
    btn: float
    cumulative: float
    inq_position: int
    is_target: int
    presented: int
    above_threshold: int

    def __iter__(self):
        return iter([getattr(self, field.name) for field in fields(self)])


def sqlite_ddl(cls: Any, table_name: str) -> str:
    """Sqlite create table statement for the given dataclass"""
    conversions = {int: 'integer', str: 'text', float: 'real'}

    column_defs = [
        f'{field.name} {conversions.get(field.type, field.type)}'
        for field in fields(cls)
    ]
    return f"CREATE TABLE {table_name} ({', '.join(column_defs)})"


def sqlite_insert(cls: Any, table_name: str) -> str:
    """sqlite INSERT statement for the given dataclass."""
    placeholders = ['?' for _ in fields(cls)]
    return f"INSERT INTO {table_name} VALUES ({','.join(placeholders)})"


def evidence_records(session: Session) -> List[EvidenceRecord]:
    """Summarize the session evidence data."""
    assert session.has_evidence(
    ), "There is no evidence in the provided session"
    assert session.symbol_set, "Session must define a symbol_set"

    records = []
    for series_index, inquiry_list in enumerate(session.series):
        for inq_index, inquiry in enumerate(inquiry_list):
            evidence = inquiry.stim_evidence(symbol_set=session.symbol_set)
            stimuli = inquiry.stimuli
            # Flatten if stimuli is structured as a nested list.
            if len(stimuli) == 1 and isinstance(stimuli[0], list):
                stimuli = stimuli[0]
            for stim, _likelihood_ev in evidence['likelihood'].items():
                records.append(
                    EvidenceRecord(
                        series=series_index + 1,
                        inquiry=inq_index,
                        stim=stim,
                        lm=evidence['lm_evidence'].get(stim, ''),
                        eeg=evidence['eeg_evidence'].get(stim, ''),
                        btn=evidence.get('btn_evidence', {}).get(stim, ''),
                        cumulative=evidence['likelihood'][stim],
                        inq_position=stimuli.index(stim) if stim in stimuli else None,
                        is_target=int(inquiry.target_letter == stim),
                        presented=int(stim in stimuli),
                        above_threshold=int(evidence['likelihood'][stim] >
                                            session.decision_threshold)))
    return records


def session_db(session: Session, db_file: str = 'session.db'):
    """Creates a sqlite database from the given session data.

    Parameters
    ----------
        session - task data (evidence values, stim times, etc.)
        db_file - path  of database to write; defaults to session.db

    Database Schema
    ---------------
    evidence:
    - trial integer (0-based)
    - inquiry integer (0-based)
    - letter text (letter or icon)
    - lm real (language model probability for the trial; same for every
        inquiry and only considered in the cumulative value during the
        first inquiry)
    - eeg real (likelihood for the given inquiry; a value of 1.0 indicates
        that the letter was not presented)
    - cumulative real (cumulative likelihood for the trial thus far)
    - inq_position integer (inquiry position; null if not presented)
    - is_target integer (boolean; true(1) if this letter is the target)
    - presented integer (boolean; true if the letter was presented in
        this inquiry)
    - above_threshold (boolean; true if cumulative likelihood was above
        the configured threshold)
    """
    # Create database
    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()
    table_name = 'evidence'
    cursor.execute(sqlite_ddl(EvidenceRecord, table_name))
    conn.commit()

    conn.executemany(sqlite_insert(EvidenceRecord, table_name),
                     map(tuple, evidence_records(session)))
    conn.commit()


def session_csv(session: Session, csv_file='session.csv'):
    """Create a csv file summarizing the evidence data for the given session."""
    with open(csv_file, "w", encoding=DEFAULT_ENCODING, newline='') as output:
        csv_writer = csv.writer(output, delimiter=',')

        columns = [field.name for field in fields(EvidenceRecord)]
        csv_writer.writerow(columns)
        for record in evidence_records(session):
            csv_writer.writerow(record)


def write_row(excel_sheet, rownum, data, background=None, border=None):
    """Helper method to write a row to an Excel spreadsheet"""
    for col, val in enumerate(data, start=1):
        cell = excel_sheet.cell(row=rownum, column=col)
        cell.value = val
        if background:
            cell.fill = background
        if border:
            cell.border = border


def session_excel(session: Session,
                  excel_file=SESSION_SUMMARY_FILENAME,
                  include_charts=True):
    """Create an Excel spreadsheet summarizing the evidence data for the given session."""

    # Define styles and borders to use within the spreadsheet.
    gray_background = PatternFill(start_color='ededed', fill_type='solid')
    white_background = PatternFill(start_color=WHITE, fill_type=None)
    highlighted_background = PatternFill(start_color=YELLOW, fill_type='solid')

    backgrounds_iter = itertools.cycle([gray_background, white_background])

    thin_gray = Side(border_style=BORDER_THIN, color='d3d3d3')
    default_border = Border(left=thin_gray,
                            top=thin_gray,
                            right=thin_gray,
                            bottom=thin_gray)

    thick_black = Side(border_style=BORDER_THIN, color=BLACK)
    new_series_border = Border(left=thin_gray,
                               top=thick_black,
                               right=thin_gray,
                               bottom=thin_gray)

    # Create the workbook
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = 'session'

    columns = [field.name for field in fields(EvidenceRecord)]
    series, inquiry = None, None

    # Write header
    write_row(sheet, 1, columns)

    # Maps the chart title to the starting row for the data.
    chart_data = {}

    # Write rows
    inq_background = next(backgrounds_iter)
    alp_len = None
    for i, record in enumerate(evidence_records(session)):
        rownum = i + 2  # Excel is 1-indexed; also account for column row.
        border = default_border

        if record.series != series:
            # Place a thick top border before each new series to make it
            # easier to visually scan the spreadsheet.
            series = record.series
            border = new_series_border

        if record.inquiry != inquiry:
            inquiry = record.inquiry

            # Set the alphabet length used for chart data ranges.
            if not alp_len and i > 0:
                alp_len = i
            chart_data[f"Series {series} inquiry {inquiry}"] = rownum

            # Toggle the background for each inquiry for easier viewing.
            inq_background = next(backgrounds_iter)

        # write to spreadsheet
        write_row(sheet,
                  rownum,
                  record,
                  background=highlighted_background
                  if bool(record.is_target) else inq_background,
                  border=border)

    # Add chart for each inquiry
    if include_charts:
        stim_col = columns.index('stim') + 1
        lm_col = columns.index('lm') + 1
        likelihood_col = columns.index('cumulative') + 1

        for title, min_row in chart_data.items():
            max_row = min_row + (alp_len - 1)
            chart = BarChart()
            chart.type = "col"
            chart.title = title
            chart.y_axis.title = 'likelihood'
            chart.x_axis.title = 'stimulus'

            data = Reference(sheet,
                             min_col=lm_col,
                             min_row=min_row,
                             max_row=max_row,
                             max_col=likelihood_col)
            categories = Reference(sheet,
                                   min_col=stim_col,
                                   min_row=min_row,
                                   max_row=max_row)
            chart.add_data(data, titles_from_data=False)
            chart.series[0].title = openpyxl.chart.series.SeriesLabel(v="lm")
            chart.series[1].title = openpyxl.chart.series.SeriesLabel(v="eeg")
            chart.series[2].title = openpyxl.chart.series.SeriesLabel(v="btn")
            chart.series[3].title = openpyxl.chart.series.SeriesLabel(
                v="combined")

            chart.set_categories(categories)
            sheet.add_chart(chart, f'M{min_row + 1}')

    # Freeze header row
    sheet.freeze_panes = 'A2'
    workbook.save(excel_file)
    print("Wrote output to " + excel_file)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Task Registry ; used to provide task options to the GUI and command line
tools. User defined tasks can be added to the Registry."""


# NOTE:
# In the future we may want to consider dynamically retrieving all subclasses
# of Task and use these to populate a registry. We could also provide
# functionality for bcipy users to define their own tasks and register them so
# they would appear as options in the GUI.
#
# However, this approach is currently problematic for the GUI interface. Due
# to the tight coupling of the display code with the Tasks, importing any of
# the Task subclasses causes pygame (a psychopy dependency) to create a GUI,
# which seems to prevent our other GUI code from working.

from enum import Enum
from typing import List

from bcipy.helpers.exceptions import BciPyCoreException


class TaskType(Enum):
    """Enum of the registered experiment types (Tasks), along with the label
    used for display in the GUI and command line tools. Values are looked up
    by their (1-based) index.

    Examples:
    >>> TaskType(1)
    <TaskType.RSVP_CALIBRATION: 1>

    >>> TaskType(1).label
    'RSVP Calibration'
    """

    RSVP_CALIBRATION = 'RSVP Calibration'
    RSVP_COPY_PHRASE = 'RSVP Copy Phrase'
    RSVP_TIMING_VERIFICATION_CALIBRATION = 'RSVP Time Test Calibration'
    MATRIX_CALIBRATION = 'Matrix Calibration'
    MATRIX_TIMING_VERIFICATION_CALIBRATION = 'Matrix Time Test Calibration'
    MATRIX_COPY_PHRASE = 'Matrix Copy Phrase'

    def __new__(cls, *args, **kwds):
        """Autoincrements the value of each item added to the enum."""
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, label):
        self.label = label

    @classmethod
    def by_value(cls, item):
        tasks = cls.list()
        # The cls.list method returns a sorted list of enum tasks
        # check if item present and return the index + 1 (which is the ENUM value for the task)
        if item in tasks:
            return cls(tasks.index(item) + 1)
        raise BciPyCoreException(f'{item} not a registered TaskType={tasks}')

    @classmethod
    def calibration_tasks(cls) -> List['TaskType']:
        return [task for task in cls
                if task.name.endswith('CALIBRATION') and 'COPY_PHRASE'
                not in task.name]

    @classmethod
    def list(cls) -> List[str]:
        return list(map(lambda c: c.label, cls))



import unittest
from bcipy.helpers.system_utils import (
    is_connected,
    is_battery_powered,
    is_screen_refresh_rate_low,
    get_screen_info,
    ScreenInfo,
)
import socket
import psutil
from pyglet import canvas
from mockito import any, when, mock, verify


class TestSystemUtilsAlerts(unittest.TestCase):

    def test_is_connected_true(self):
        """Test that a computer connected to the internet returns True."""
        mock_conn = mock()
        when(socket).create_connection(address=any, timeout=any).thenReturn(mock_conn)
        self.assertTrue(is_connected())

    def test_is_connected_false(self):
        """Test that a computer not connected to the internet returns False."""
        when(socket).create_connection(address=any, timeout=any).thenRaise(OSError)
        self.assertFalse(is_connected())

    def test_is_battery_powered_true(self):
        """Test that a computer running on battery returns True."""
        mock_battery = mock()
        mock_battery.power_plugged = False
        when(psutil).sensors_battery().thenReturn(mock_battery)
        self.assertTrue(is_battery_powered())

    def test_is_battery_powered_false(self):
        """Test that a computer plugged in returns False."""
        mock_battery = mock()
        mock_battery.power_plugged = True
        when(psutil).sensors_battery().thenReturn(mock_battery)
        self.assertFalse(is_battery_powered())

    def test_is_screen_refresh_rate_low(self):
        """Test that a refresh rate of less than 120 Hz returns True."""
        display = mock()
        screen = mock()
        screen.width = 1920
        screen.height = 1080
        screen.rate = 60
        when(canvas).get_display().thenReturn(display)
        when(display).get_default_screen().thenReturn(screen)
        when(screen).get_mode().thenReturn(screen)

        self.assertTrue(is_screen_refresh_rate_low())
        verify(display, times=1).get_default_screen()
        verify(screen, times=1).get_mode()

    def test_is_screen_refresh_rate_low_false(self):
        """Test that a refresh rate of 120 Hz or greater returns False."""
        display = mock()
        screen = mock()
        screen.width = 1920
        screen.height = 1080
        screen.rate = 120
        when(canvas).get_display().thenReturn(display)
        when(display).get_default_screen().thenReturn(screen)
        when(screen).get_mode().thenReturn(screen)
        self.assertFalse(is_screen_refresh_rate_low())
        verify(display, times=1).get_default_screen()
        verify(screen, times=1).get_mode()

    def test_is_screen_refresh_ok_given_refresh(self):
        """Test that a refresh rate of 120 Hz or greater returns False."""
        self.assertFalse(is_screen_refresh_rate_low(120))

    def test_get_screen_info(self):
        """Test that the screen info is returned as expected."""
        display = mock()
        screen = mock()
        screen.width = 1920
        screen.height = 1080
        screen.rate = 120
        when(canvas).get_display().thenReturn(display)
        when(display).get_default_screen().thenReturn(screen)
        when(screen).get_mode().thenReturn(screen)
        expected = ScreenInfo(width=1920, height=1080, rate=120)
        self.assertEqual(get_screen_info(), expected)
        verify(display, times=1).get_default_screen()
        verify(screen, times=1).get_mode()


if __name__ == "__main__":
    unittest.main()



"""Defines the CopyPhraseWrapper."""
from typing import List, Tuple

import logging
import numpy as np
from bcipy.helpers.acquisition import analysis_channels, DeviceSpec
from bcipy.helpers.exceptions import BciPyCoreException
from bcipy.helpers.language_model import (
    histogram,
    with_min_prob,
)
from bcipy.helpers.stimuli import InquirySchedule, StimuliOrder, TrialReshaper
from bcipy.helpers.symbols import BACKSPACE_CHAR
from bcipy.signal.model import SignalModel
from bcipy.signal.process import get_default_transform
from bcipy.task.control.handler import DecisionMaker, EvidenceFusion
from bcipy.task.control.query import NBestStimuliAgent
from bcipy.task.control.criteria import (
    CriteriaEvaluator,
    MaxIterationsCriteria,
    MinIterationsCriteria,
    ProbThresholdCriteria,
)
from bcipy.task.data import EvidenceType
from bcipy.language.main import LanguageModel


log = logging.getLogger(__name__)


class CopyPhraseWrapper:
    """Basic copy phrase task duty cycle wrapper. Coordinates activities around
    spelling tasks, including:

    - Evidence management: adding and fusing evidence using the EvidenceFusion
    module.
    - Decision-making: uses evidence to make a decision or decide to continue
    by providing another inquiry (DecisionMaker).
    - Determining when to stop an inquiry and make a decision (StoppageCriteria).
    - Generation of inquiries.
    - Coordination with the Language Model.
    - Preparing EEG data for the SignalModel for classification.

    Parameters
    ----------
    - min_num_inq: The minimum number of inquiries to be displayed
    - max_num_inq: The maximum number of inquiries to be displayed
    - device_spec: Specification for the EEG device used for data collection.
    - signal_model: model trained using a calibration session of the same user.
    - k: down sampling rate
    - alp: symbol set of the task
    - evidence_names: list of evidence types used for decision-making
    - task_list: list[(phrases, initial_states)] for the copy phrase task
    - lmodel: language model used (when 'LM') evidence type is used.
    - is_txt_stim: Whether or not the stimuli are text objects
    - decision_threshold: Minimum likelihood value required for a decision
    - backspace_prob: default language model probability for the
    backspace character.
    - backspace_always_shown: whether or not the backspace should
    always be presented.
    - filter_high: filter setting used when evaluating EEG data
    - filter_low: filter setting used when evaluating EEG data
    - filter_order: filter setting used when evaluating EEG data
    - notch_filter_frequency: filter setting used when evaluating EEG data
    - stim_length(int): the number of stimuli to present in each inquiry
    - stim_timing: seconds each stimuli is displayed; used for inquiry
    generation
    - stim_jitter: seconds that inquiry stimuli should be jittered (-/+ stim_timing[-1])
    """

    def __init__(self,
                 min_num_inq: int,
                 max_num_inq: int,
                 lmodel: LanguageModel,
                 device_spec: DeviceSpec,
                 signal_model: SignalModel = None,
                 k: int = 2,
                 alp: List[str] = None,
                 evidence_names: List[EvidenceType] = [
                     EvidenceType.LM, EvidenceType.ERP
                 ],
                 task_list: List[Tuple[str, str]] = [('I_LOVE_COOKIES',
                                                      'I_LOVE_')],
                 is_txt_stim: bool = True,
                 decision_threshold: float = 0.8,
                 backspace_prob: float = 0.05,
                 backspace_always_shown: bool = False,
                 filter_high: int = 45,
                 filter_low: int = 2,
                 filter_order: int = 2,
                 notch_filter_frequency: int = 60,
                 stim_timing: List[float] = [1, .2],
                 stim_length: int = 10,
                 stim_jitter: float = 0,
                 stim_order: StimuliOrder = StimuliOrder.RANDOM):

        self.lmodel = lmodel
        self.conjugator = EvidenceFusion(evidence_names, len_dist=len(alp))

        inq_constants = []
        if backspace_always_shown and BACKSPACE_CHAR in alp:
            inq_constants.append(BACKSPACE_CHAR)

        # Stimuli Selection Module
        stopping_criteria = CriteriaEvaluator(
            continue_criteria=[MinIterationsCriteria(min_num_inq)],
            commit_criteria=[MaxIterationsCriteria(max_num_inq),
                             ProbThresholdCriteria(decision_threshold)])

        self.stim_length = stim_length
        self.stim_order = stim_order
        stimuli_agent = NBestStimuliAgent(alphabet=alp,
                                          len_query=self.stim_length)

        self.decision_maker = DecisionMaker(
            stimuli_agent=stimuli_agent,
            stopping_evaluator=stopping_criteria,
            state=task_list[0][1],
            alphabet=alp,
            is_txt_stim=is_txt_stim,
            stimuli_jitter=stim_jitter,
            stimuli_timing=stim_timing,
            stimuli_order=self.stim_order,
            inq_constants=inq_constants)

        self.alp = alp
        # non-letter target labels include the fixation cross and calibration.
        self.nonletters = ['+', 'PLUS', 'calibration_trigger']
        self.valid_targets = set(self.alp)

        self.signal_model = signal_model
        self.sampling_rate = device_spec.sample_rate
        self.downsample_rate = k
        self.filter_high = filter_high
        self.filter_low = filter_low
        self.filter_order = filter_order
        self.notch_filter_frequency = notch_filter_frequency

        self.mode = 'copy_phrase'
        self.task_list = task_list
        self.channel_map = analysis_channels(device_spec.channels, device_spec)
        self.backspace_prob = backspace_prob

    def evaluate_inquiry(
            self, raw_data: np.array, triggers: List[Tuple[str, float]],
            target_info: List[str], window_length: float
    ) -> Tuple[bool, Tuple[List[str], List[float], List[str]]]:
        """Once data is collected, infers meaning from the data and attempt to
        make a decision.

        Parameters
        ----------
        - raw_data: C x L eeg data where C is number of channels and L is the
        signal length
        - triggers: triggers e.g. `('A', 1)` as letter and flash time for the
        letter
        - target_info: target information about the stimuli
        - window_length: The length of the time between stimuli presentation

        Returns
        -------
        - (True, None) when commitment is made.
        - (False, next set of stimuli) when not enough evidence has
        been provided and stoppage criteria is not yet met.
        """
        lik_r = self.evaluate_eeg_evidence(raw_data, triggers, target_info,
                                           window_length)
        self.add_evidence(EvidenceType.ERP, lik_r)
        return self.decide()

    def evaluate_eeg_evidence(self, raw_data: np.array,
                              triggers: List[Tuple[str, float]],
                              target_info: List[str],
                              window_length: float) -> np.array:
        """Once data is collected, infers meaning from the data and return the results.

        Parameters
        ----------
        - raw_data: C x L eeg data where C is number of channels and L is the
        signal length
        - triggers: triggers e.g. `('A', 1)` as letter and flash time for the
        letter
        - target_info: target information about the stimuli
        - window_length: The length of the time between stimuli presentation

        Returns
        -------
        np.array of likelihood evidence
        """
        letters, times, target_info = self.letter_info(triggers, target_info)

        default_transform = get_default_transform(
            sample_rate_hz=self.sampling_rate,
            notch_freq_hz=self.notch_filter_frequency,
            bandpass_low=self.filter_low,
            bandpass_high=self.filter_high,
            bandpass_order=self.filter_order,
            downsample_factor=self.downsample_rate,
        )

        data, transformed_sample_rate = default_transform(raw_data, self.sampling_rate)

        # The data from DAQ is assumed to have offsets applied
        data, _ = TrialReshaper()(
            trial_targetness_label=target_info,
            timing_info=times,
            eeg_data=data,
            sample_rate=transformed_sample_rate,
            channel_map=self.channel_map,
            poststimulus_length=window_length)

        return self.signal_model.predict(data, letters, self.alp)

    def add_evidence(self, evidence_type: EvidenceType,
                     evidence: List[float]) -> np.array:
        """Add evidence to the conjugator.

        Parameters
        ----------
        - evidence_type : type of evidence (ex. `'LM'`, `'ERP'`)
        - evidence : ndarray[float], evidence for each stim

        Returns
        -------
        updated likelihoods after fusing the new evidence
        """
        assert evidence_type in self.conjugator.evidence_history.keys(
        ), f"Copy Phrase wrapper was not initialized with evidence type: {evidence_type}."
        return self.conjugator.update_and_fuse(
            {evidence_type: np.array(evidence)})

    def decide(self) -> Tuple[bool, InquirySchedule]:
        """Make a decision based on the current evidence.

        Returns
        -------
        - (True, None) when commitment is made.
        - (False, next set of stimuli) when not enough evidence has
        been provided and stoppage criteria is not yet met.
        """
        decision, new_stim = self.decision_maker.decide(
            self.conjugator.likelihood[:])
        return decision, new_stim

    def letter_info(self, triggers: List[Tuple[str, float]],
                    target_info: List[str]
                    ) -> Tuple[List[str], List[float], List[str]]:
        """
        Filters out non-letters and separates timings from letters.

        Parameters
        ----------
        - triggers: triggers e.g. [['A', 0.5], ...]
        as letter and flash time for the letter
        - target_info: target information about the stimuli;
        ex. ['nontarget', 'nontarget', ...]

        Returns
        -------
        (letters, times, target_info)
        """
        letters = []
        times = []
        target_types = []

        for i, (letter, stamp) in enumerate(triggers):
            if letter not in self.nonletters:
                letters.append(letter)
                times.append(stamp)
                target_types.append(target_info[i])

        # Raise an error if the stimuli includes unexpected terms
        if not set(letters).issubset(self.valid_targets):
            invalid = set(letters).difference(self.valid_targets)
            error_message = f'unexpected letters received in copy phrase: {invalid}'
            log.error(error_message)
            raise BciPyCoreException(error_message)

        return letters, times, target_types

    def initialize_series(self) -> Tuple[bool, InquirySchedule]:
        """If a decision is made initializes the next series."""
        assert self.lmodel, "Language model must be initialized."

        try:
            # First, reset the history for this new series
            self.conjugator.reset_history()

            # Get the displayed state
            update = self.decision_maker.displayed_state
            log.info(f"Querying language model: '{update}'")

            # update the lmodel and get back the priors
            lm_letter_prior = self.lmodel.predict(list(update))

            if BACKSPACE_CHAR in self.alp:
                # Apply configured backspace probability.
                sym = (BACKSPACE_CHAR, self.backspace_prob)
                lm_letter_prior = with_min_prob(lm_letter_prior, sym)

            # convert to format needed for evidence fusion;
            # probability value only in alphabet order.
            prior = [
                prior_prob for alp_letter in self.alp
                for prior_sym, prior_prob in lm_letter_prior
                if alp_letter == prior_sym
            ]

            # display histogram of LM probabilities
            log.debug(histogram(lm_letter_prior))

            # Try fusing the lmodel evidence
            try:
                prob_dist = self.conjugator.update_and_fuse(
                    {EvidenceType.LM: np.array(prior)})
            except Exception as fusion_error:
                log.exception(f'Error fusing language model evidence!: {fusion_error}')
                raise BciPyCoreException(fusion_error) from fusion_error

            # Get decision maker to give us back some decisions and stimuli
            is_accepted, sti = self.decision_maker.decide(prob_dist)

        except Exception as init_series_error:
            log.exception(f'Error in initialize_series: {init_series_error}')
            raise BciPyCoreException(init_series_error) from init_series_error

        return is_accepted, sti



import numpy as np
from abc import ABC, abstractmethod


class Rule(ABC):
    """Rule.

    Python Abstract Base Class for a Rule.
    (https://docs.python.org/3/library/abc.html)
    Each rule has a 'is_broken' method which acts as instructions for
    the evaluator to use when evaluating data. Returns True upon
    rule breakage, otherwise defaults to False."""

    def __init__(self, threshold):
        self.threshold = threshold

    # a method required for all subclasses
    @abstractmethod
    def is_broken(self, data):
        ...


class HighVoltage(Rule):
    """High Voltage Rule.

    Set high threshold for permitted voltage.

    Allows different types of rules to be fed to artifact rejector so that experimenter can differentiate between
    high and low warnings easily. Makes it so that rules are singularly defined. Names must be equal to value in
    parameters.json
    """

    def is_broken(self, data):
        """Is Broken.

            Test data against threshold value. Return broken
            if threshold exceeded.

            data(ndarray[float]): C x N length array where
            C is the number of channels and N is the number of samples

                np.amax takes the maximum value in an
                array even of length 1:
                (https://docs.scipy.org/doc/numpy/
                reference/generated/numpy.amax.html)
        """
        if np.amax(data) >= self.threshold:
            return True

        return False

    def __str__(self):
        return f'High Voltage with threshold {self.threshold}'


class LowVoltage(Rule):
    """Low Voltage Rule.

    Set low threshold for permitted voltage.
    """

    def is_broken(self, data):
        """Is Broken.

            Test data against threshold value. Return false
            if threshold exceeded.


            data(ndarray[float]): C x N length array where
            C is the number of channels and N is the number of samples


                np.amin takes the minimum value in an array even of length 1:
                  (https://docs.scipy.org/doc/numpy/reference/generated/numpy.amin.html)
         """
        if np.amin(data) <= self.threshold:
            return True

        return False

    def __str__(self):
        return f'Low Voltage with threshold {self.threshold}'



import unittest

from typing import List
from collections import Counter
from mockito import unstub, mock, when, verify, verifyStubbedInvocationsAreUsed

import numpy as np
import psychopy

from bcipy.acquisition import LslAcquisitionClient
from bcipy.acquisition.record import Record
from bcipy.task.exceptions import InsufficientDataException

from bcipy.helpers.task import (_float_val,
                                calculate_stimulation_freq, construct_triggers,
                                generate_targets, get_data_for_decision,
                                get_key_press, target_info)


class TestCalculateStimulationFreq(unittest.TestCase):
    def test_calculate_stimulate_frequency_returns_number_less_one(self):
        flash_time = 5
        stimulation_frequency = calculate_stimulation_freq(flash_time)
        expected = 1 / flash_time
        self.assertEqual(stimulation_frequency, expected)

    def test_calculate_stimulate_frequency_handles_zero(self):
        flash_time = 0
        with self.assertRaises(ZeroDivisionError):
            calculate_stimulation_freq(flash_time)


class TestFloatVal(unittest.TestCase):
    def test_float_val_as_str(self):
        col = 'Apple'
        result = _float_val(col)
        expected = 1.0
        self.assertEqual(result, expected)

    def test_float_val_as_int(self):
        col = 3
        result = _float_val(col)
        expected = 3.0
        self.assertEqual(result, expected)


class TestTargetGeneration(unittest.TestCase):
    """Tests for generation of target inquiries"""

    def test_target_number_less_than_alp(self):
        """Test when requested number of targets is less than the length of
        the alphabet."""
        alp = list(range(10))
        targets = generate_targets(alp, 5)
        self.assertEqual(len(targets), 5)
        self.assertEqual(len(targets), len(set(targets)))

    def test_target_greater_than_alp(self):
        """Test behavior when number of targets is greater than the length
        of the alphabet"""
        alp = list(range(5))
        targets = generate_targets(alp, 10)
        self.assertEqual(len(targets), 10)

        counts = Counter(targets)

        for item in alp:
            self.assertEqual(counts[item], 2)

    def test_remainder(self):
        """Test behavior when number of targets is greater than the length of
        the alphabet by a value other than a multiple of the alphabet length.
        """
        alp = list(range(5))
        targets = generate_targets(alp, 12)

        counts = Counter(targets)
        for item in alp:
            self.assertGreaterEqual(counts[item], 2)
            self.assertLessEqual(counts[item], 3)


class TestGetKeyPress(unittest.TestCase):
    """Tests for the get key press method"""

    def tearDown(self):
        verifyStubbedInvocationsAreUsed()
        unstub()

    def test_get_key_press_appends_stamp_label_defaults(self):
        """Test for the stamp label defaults, ensures the calls occur with the correct inputs to psychopy"""
        key_list = ['space']
        clock = mock()
        # get keys returns a list of lists with the key and timestamp per hit
        key_response = [[key_list[0], 1000]]
        when(psychopy.event).getKeys(keyList=key_list,
                                     timeStamped=True).thenReturn(key_response)
        when(clock).getTime().thenReturn(psychopy.core.getTime())
        # use the default label
        stamp_label = 'bcipy_key_press'
        expected = [f'{stamp_label}_{key_response[0][0]}', key_response[0][1]]
        response = get_key_press(key_list, clock)
        self.assertEqual(expected[0], response[0])
        self.assertAlmostEqual(expected[1], response[1], delta=0.01)

    def test_get_key_press_clock_adjustment(self):
        """Test for the stamp label defaults, ensures the calls occur with the correct inputs to psychopy"""
        key_list = ['space']
        clock = mock()
        # get keys returns a list of lists with the key and timestamp per hit
        key_response = [[key_list[0], 1000]]
        when(psychopy.event).getKeys(keyList=key_list,
                                     timeStamped=True).thenReturn(key_response)
        when(clock).getTime().thenReturn(psychopy.core.getTime() + 100)
        # use the default label
        stamp_label = 'bcipy_key_press'
        expected = [f'{stamp_label}_{key_response[0][0]}', key_response[0][1]]
        response = get_key_press(key_list, clock)
        self.assertEqual(expected[0], response[0])
        self.assertAlmostEqual(1100, response[1], delta=0.01)

    def test_get_key_press_returns_none_if_no_keys_pressed(self):
        """Test for the case not keys are returned, ensures the calls occur with the correct inputs to psychopy"""

        key_list = ['space']
        key_response = None
        clock = mock()
        when(psychopy.event).getKeys(keyList=key_list,
                                     timeStamped=True).thenReturn(key_response)

        response = get_key_press(key_list, clock)
        self.assertEqual(None, response)

    def test_get_key_press_set_custom_stamp_message(self):
        """Test for a custom stamp label, ensures the calls occur with the correct inputs to psychopy"""
        clock = mock()
        key_list = ['space']
        # get keys returns a list of lists with the key and timestamp per hit
        key_response = [[key_list[0], 1000]]
        when(psychopy.event).getKeys(keyList=key_list,
                                     timeStamped=True).thenReturn(key_response)
        when(clock).getTime().thenReturn(psychopy.core.getTime())
        # set a custom label
        stamp_label = 'custom_label'
        expected = [f'{stamp_label}_{key_response[0][0]}', key_response[0][1]]
        response = get_key_press(key_list, clock, stamp_label=stamp_label)
        self.assertEqual(expected[0], response[0])
        self.assertAlmostEqual(expected[1], response[1], delta=0.01)


class TestTriggers(unittest.TestCase):
    """Tests related to triggers"""

    def test_construct_triggers(self):
        stim_times = [['+', 7.009946188016329], ['<', 7.477798109990545],
                      ['_', 7.69470399999409], ['Z', 7.911495972017292],
                      ['U', 8.128477902995655], ['S', 8.345279764995212],
                      ['T', 8.562265532993479], ['V', 8.779025560012087],
                      ['X', 8.995945784990909], ['Y', 9.213076218002243],
                      ['W', 9.429835998016642]]
        expected = [('+', 0.0), ('<', 0.46785192197421566),
                    ('_', 0.6847578119777609), ('Z', 0.901549784000963),
                    ('U', 1.1185317149793264), ('S', 1.3353335769788828),
                    ('T', 1.5523193449771497), ('V', 1.7690793719957583),
                    ('X', 1.9859995969745796), ('Y', 2.203130029985914),
                    ('W', 2.4198898100003134)]
        self.assertEqual(expected, construct_triggers(stim_times))
        self.assertEqual([], construct_triggers([]))

    def test_target_info(self):
        triggers = [('+', 0.0), ('<', 0.46785192197421566),
                    ('_', 0.6847578119777609), ('Z', 0.901549784000963),
                    ('U', 1.1185317149793264), ('S', 1.3353335769788828),
                    ('T', 1.5523193449771497), ('V', 1.7690793719957583),
                    ('X', 1.9859995969745796), ('Y', 2.203130029985914),
                    ('W', 2.4198898100003134)]
        expected = [
            'fixation', 'nontarget', 'nontarget', 'target', 'nontarget',
            'nontarget', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget'
        ]
        self.assertEqual(expected, target_info(triggers, target_letter='Z'))

        expected = [
            'fixation', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget'
        ]
        self.assertEqual(expected, target_info(triggers))
        self.assertEqual([], target_info([]))


def mock_get_data_response(samples: int, high: float, low: float,
                           channels: int) -> List[Record]:
    """Mock DataAcquisitionClient Response.

    The data acquisition client returns a list of records that need to be looped through
        to get the raw data without other items attached.
    """
    data = [np.random.uniform(low, high) for _ in range(channels)]
    record_data = []
    for i in range(samples):
        record_data.append(Record(data, i, None))
    return record_data


class TestGetDataForDecision(unittest.TestCase):
    def setUp(self) -> None:
        self.inquiry_timing = [('A', 1), ('B', 2), ('C', 3)]
        self.daq = mock(spec=LslAcquisitionClient)
        self.daq.device_spec = mock()
        self.daq.device_spec.sample_rate = 10
        self.mock_eeg = mock_get_data_response(samples=1000,
                                               high=1000,
                                               low=-1000,
                                               channels=4)

    def tearDown(self) -> None:
        unstub()

    def test_get_data_for_decision_returns_tuple_of_eeg_data_and_triggers(
            self):
        when(self.daq).get_data(start=any, limit=any).thenReturn(self.mock_eeg)

        response = get_data_for_decision(self.inquiry_timing, self.daq)

        self.assertIsInstance(response, tuple)

        _eeg_data, timing = response

        # self.assertEqual(eeg_data[:1][0], self.mock_eeg[0].data[0])
        self.assertIsInstance(timing, list)

    def test_get_data_for_decision_prestim(self):
        prestim = 1
        first_stim_time = self.inquiry_timing[0][1]
        last_stim_time = self.inquiry_timing[-1][1]

        expected_start = first_stim_time - prestim
        expected_stop = last_stim_time
        expected_triggers = [(text, ((timing - first_stim_time) + prestim))
                             for text, timing in self.inquiry_timing]
        expected_data_limit = round((expected_stop - expected_start) *
                                    self.daq.device_spec.sample_rate)

        when(self.daq).get_data(start=expected_start,
                                limit=expected_data_limit).thenReturn(
                                    self.mock_eeg)
        _, timing = get_data_for_decision(self.inquiry_timing,
                                          self.daq,
                                          prestim=prestim)

        # self.assertEqual(eeg_data[:1][0], self.mock_eeg[0].data[0])
        self.assertEqual(timing, expected_triggers)
        verify(self.daq, times=1).get_data(start=expected_start,
                                           limit=expected_data_limit)

    def test_get_data_for_decision_poststim(self):
        poststim = 1
        first_stim_time = self.inquiry_timing[0][1]
        last_stim_time = self.inquiry_timing[-1][1]

        expected_triggers = [(text, ((timing) - first_stim_time))
                             for text, timing in self.inquiry_timing]
        expected_data_limit = round(
            (last_stim_time - first_stim_time + poststim) *
            self.daq.device_spec.sample_rate)

        when(self.daq).get_data(start=first_stim_time,
                                limit=expected_data_limit).thenReturn(
                                    self.mock_eeg)
        _, timing = get_data_for_decision(self.inquiry_timing,
                                          self.daq,
                                          poststim=poststim)

        # self.assertEqual(eeg_data[:1][0], self.mock_eeg[0].data[0])
        self.assertEqual(timing, expected_triggers)
        verify(self.daq, times=1).get_data(start=first_stim_time,
                                           limit=expected_data_limit)

    def test_get_data_for_decision_offset(self):
        offset = 1
        first_stim_time = self.inquiry_timing[0][1]
        last_stim_time = self.inquiry_timing[-1][1]

        expected_start = first_stim_time + offset
        expected_stop = last_stim_time + offset
        expected_triggers = [(text, ((timing) - first_stim_time))
                             for text, timing in self.inquiry_timing]
        expected_data_limit = round((expected_stop - expected_start) *
                                    self.daq.device_spec.sample_rate)

        when(self.daq).get_data(start=expected_start,
                                limit=expected_data_limit).thenReturn(
                                    self.mock_eeg)
        _, timing = get_data_for_decision(self.inquiry_timing,
                                          self.daq,
                                          offset=offset)

        # self.assertEqual(eeg_data[:1][0], self.mock_eeg[0].data[0])
        self.assertEqual(timing, expected_triggers)
        verify(self.daq, times=1).get_data(start=expected_start,
                                           limit=expected_data_limit)

    def test_get_data_for_decision_throws_insufficient_data_error_if_less_than_data_limit(
            self):

        # return an empty list from the get data call
        when(self.daq).get_data(start=any, limit=any).thenReturn([])

        with self.assertRaises(InsufficientDataException):
            get_data_for_decision(self.inquiry_timing, self.daq)

    def test_get_data_for_decision_throws_insufficient_data_error_if_data_query_out_of_bounds(
            self):
        inquiry_timing = [('A', 10), ('D', 1)]

        with self.assertRaises(InsufficientDataException):
            get_data_for_decision(inquiry_timing, self.daq)


if __name__ == '__main__':
    unittest.main()



from psychopy import core
from typing import List, Tuple

from bcipy.config import TRIGGER_FILENAME, WAIT_SCREEN_MESSAGE
from bcipy.display import (
    InformationProperties,
    PreviewInquiryProperties,
    StimuliProperties)
from bcipy.display.paradigm.rsvp.mode.calibration import CalibrationDisplay
from bcipy.display.components.task_bar import CalibrationTaskBar
from bcipy.helpers.clock import Clock

from bcipy.helpers.stimuli import (StimuliOrder, TargetPositions, calibration_inquiry_generator)
from bcipy.helpers.task import (get_user_input, pause_calibration,
                                trial_complete_message)
from bcipy.helpers.symbols import alphabet
from bcipy.helpers.triggers import FlushFrequency, TriggerHandler, Trigger, TriggerType, convert_timing_triggers
from bcipy.task import Task


class RSVPCalibrationTask(Task):
    """RSVP Calibration Task.

    Calibration task performs an RSVP stimulus inquiry
        to elicit an ERP. Parameters will change how many stimuli
        and for how long they present. Parameters also change
        color and text / image inputs.

    This task progresses as follows:

    setting up variables --> initializing eeg --> awaiting user input to start --> setting up stimuli -->
    presenting inquiries --> saving data

    PARAMETERS:
    ----------
    win (PsychoPy Display)
    daq (Data Acquisition Client)
    parameters (dict)
    file_save (str)
    """

    def __init__(self, win, daq, parameters, file_save):
        super(RSVPCalibrationTask, self).__init__()

        self.window = win
        self.frame_rate = self.window.getActualFrameRate()
        self.parameters = parameters
        self.daq = daq
        self.static_clock = core.StaticPeriod(screenHz=self.frame_rate)
        self.experiment_clock = Clock()
        self.buffer_val = parameters['task_buffer_length']
        self.alp = alphabet(parameters)
        self.rsvp = init_calibration_display_task(
            self.parameters, self.window,
            self.static_clock, self.experiment_clock)
        self.file_save = file_save
        self.trigger_handler = TriggerHandler(
            self.file_save,
            TRIGGER_FILENAME,
            FlushFrequency.EVERY)

        self.stim_number = parameters['stim_number']
        self.stim_length = parameters['stim_length']
        self.stim_order = StimuliOrder(parameters['stim_order'])
        self.target_positions = TargetPositions(parameters['target_positions'])
        self.nontarget_inquiries = parameters['nontarget_inquiries']

        self.show_preview_inquiry = parameters['show_preview_inquiry']

        self.timing = [parameters['time_prompt'],
                       parameters['time_fixation'],
                       parameters['time_flash']]
        self.jitter = parameters['stim_jitter']

        self.color = [parameters['target_color'],
                      parameters['fixation_color'],
                      parameters['stim_color']]
        self.wait_screen_message_color = self.color[-1]
        self.task_info_color = parameters['task_color']

        self.stimuli_height = parameters['stim_height']
        self.is_txt_stim = parameters['is_txt_stim']
        self.enable_breaks = parameters['enable_breaks']

    def generate_stimuli(self):
        """Generates the inquiries to be presented.
        Returns:
        --------
            tuple(
                samples[list[list[str]]]: list of inquiries
                timing(list[list[float]]): list of timings
                color(list(list[str])): list of colors)
        """
        return calibration_inquiry_generator(self.alp,
                                             stim_number=self.stim_number,
                                             stim_length=self.stim_length,
                                             stim_order=self.stim_order,
                                             target_positions=self.target_positions,
                                             nontarget_inquiries=self.nontarget_inquiries,
                                             timing=self.timing,
                                             jitter=self.jitter,
                                             is_txt=self.rsvp.is_txt_stim,
                                             color=self.color)

    def trigger_type(self, symbol: str, target: str, index: int) -> TriggerType:
        """Trigger Type.

        This method is passed to convert_timing_triggers to properly assign TriggerTypes
            to the timing of stimuli presented.
        """
        if index == 0:
            return TriggerType.PROMPT
        if symbol == 'inquiry_preview' and index == 1:
            return TriggerType.PREVIEW
        if symbol == '+':
            return TriggerType.FIXATION
        if target == symbol:
            return TriggerType.TARGET
        return TriggerType.NONTARGET

    def execute(self):

        self.logger.debug(f'Starting {self.name()}!')
        run = True

        # Check user input to make sure we should be going
        if not get_user_input(self.rsvp, WAIT_SCREEN_MESSAGE,
                              self.wait_screen_message_color,
                              first_run=True):
            run = False

        # Begin the Experiment
        while run:

            # Get inquiry information given stimuli parameters
            (stimuli, stimuli_timing, stimuli_color) = self.generate_stimuli()

            # Execute the RSVP inquiries
            for inquiry in range(self.stim_number):

                # check user input to make sure we should be going
                if not get_user_input(self.rsvp, WAIT_SCREEN_MESSAGE,
                                      self.wait_screen_message_color):
                    break

                # Take a break every number of trials defined
                if self.enable_breaks:
                    pause_calibration(self.window, self.rsvp, inquiry,
                                      self.parameters)

                # update task state
                self.rsvp.update_task_bar()

                # Draw and flip screen
                self.rsvp.draw_static()
                self.window.flip()

                # Schedule a inquiry
                self.rsvp.stimuli_inquiry = stimuli[inquiry]

                # check if text stimuli or not for color information
                if self.is_txt_stim:
                    self.rsvp.stimuli_colors = stimuli_color[inquiry]

                self.rsvp.stimuli_timing = stimuli_timing[inquiry]

                core.wait(self.buffer_val)

                # do inquiry and get timing information back from the display
                timing = self.rsvp.do_inquiry(preview_calibration=self.show_preview_inquiry)

                self.write_trigger_data(timing, (inquiry == 0))
                core.wait(self.buffer_val)

            # Set run to False to stop looping
            run = False

        # Say Goodbye!
        self.rsvp.info_text = trial_complete_message(self.window, self.parameters)
        self.rsvp.draw_static()
        self.window.flip()

        self.write_offset_trigger()

        # Wait some time before exiting so there is trailing eeg data saved
        core.wait(self.buffer_val)

        return self.file_save

    def write_trigger_data(self, timing: List[Tuple[str, float]], first_run) -> None:
        """Write Trigger Data.

        Using the timing provided from the display and calibration information from the data acquisition
        client, write trigger data in the correct format.

        *Note on offsets*: we write the full offset value which can be used to transform all stimuli to the time since
            session start (t = 0) for all values (as opposed to most system clocks which start much higher).
            We do not write the calibration trigger used to generate this offset from the display.
            See RSVPDisplay._trigger_pulse() for more information.
        """
        # write offsets. currently, we only check for offsets at the beginning.
        if self.daq.is_calibrated and first_run:
            self.trigger_handler.add_triggers(
                [Trigger(
                    'starting_offset',
                    TriggerType.OFFSET,
                    # offset will factor in true offset and time relative from beginning
                    (self.daq.offset(self.rsvp.first_stim_time) - self.rsvp.first_stim_time)
                )]
            )

        # make sure triggers are written for the inquiry
        self.trigger_handler.add_triggers(convert_timing_triggers(timing, timing[0][0], self.trigger_type))

    def write_offset_trigger(self) -> None:
        """Append an offset value to the end of the trigger file.
        """
        if self.daq.is_calibrated:
            self.trigger_handler.add_triggers(
                [Trigger(
                    'daq_sample_offset',
                    TriggerType.SYSTEM,
                    # to help support future refactoring or use of lsl timestamps only
                    # we write only the sample offset here
                    self.daq.offset(self.rsvp.first_stim_time)
                )])
        self.trigger_handler.close()

    def name(self):
        return 'RSVP Calibration Task'


def init_calibration_display_task(
        parameters, window, static_clock, experiment_clock):
    info = InformationProperties(
        info_color=[parameters['info_color']],
        info_pos=[(parameters['info_pos_x'],
                   parameters['info_pos_y'])],
        info_height=[parameters['info_height']],
        info_font=[parameters['font']],
        info_text=[parameters['info_text']],
    )
    stimuli = StimuliProperties(stim_font=parameters['font'],
                                stim_pos=(parameters['stim_pos_x'],
                                          parameters['stim_pos_y']),
                                stim_height=parameters['stim_height'],
                                stim_inquiry=[''] * parameters['stim_length'],
                                stim_colors=[parameters['stim_color']] * parameters['stim_length'],
                                stim_timing=[10] * parameters['stim_length'],
                                is_txt_stim=parameters['is_txt_stim'])

    task_bar = CalibrationTaskBar(window,
                                  inquiry_count=parameters['stim_number'],
                                  current_index=0,
                                  colors=[parameters['task_color']],
                                  font=parameters['font'],
                                  height=parameters['task_height'])

    preview_inquiry = PreviewInquiryProperties(
        preview_only=True,
        preview_inquiry_length=parameters['preview_inquiry_length'],
        preview_inquiry_progress_method=parameters['preview_inquiry_progress_method'],
        preview_inquiry_key_input=parameters['preview_inquiry_key_input'],
        preview_inquiry_isi=parameters['preview_inquiry_isi'])

    return CalibrationDisplay(
        window,
        static_clock,
        experiment_clock,
        stimuli,
        task_bar,
        info,
        preview_inquiry=preview_inquiry,
        trigger_type=parameters['trigger_type'],
        space_char=parameters['stim_space_char'],
        full_screen=parameters['full_screen'])



"""Tests for BciPy raw data format."""
import os
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from mockito import any, mock, when, verify, unstub

from bcipy.helpers.exceptions import BciPyCoreException
from bcipy.helpers.raw_data import (RawData, RawDataReader, RawDataWriter,
                                    load, sample_data, settings, write)


class TestRawData(unittest.TestCase):
    """Tests for raw_data format"""

    def setUp(self):
        """Override; set up the needed path for load functions."""
        self.data_dir = f"{os.path.dirname(__file__)}/resources/"
        self.temp_dir = tempfile.mkdtemp()
        self.path = Path(self.temp_dir, 'test_raw_data.csv')
        self.daq_type = 'Test-Device'
        self.sample_rate = 300.0
        self.columns = ['timestamp', 'ch1', 'ch2', 'ch3']
        self.row1 = [1, 1.0, 2.0, 3.0]
        self.row2 = [2, 4.0, 5.0, 6.0]
        self.row3 = [3, 7.0, 8.0, 9.0]

    def tearDown(self):
        """Override"""
        shutil.rmtree(self.temp_dir)
        unstub()

    def _write_raw_data(self, include_rows=False) -> RawData:
        """Helper function to write a sample raw data file to disk using the
        settings from setUp.

        Parameters
        ----------
        - include_rows : if True adds data, otherwise just writes the metadata
        and columns.
        """
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=self.columns)
        if include_rows:
            data.append(self.row1)
            data.append(self.row2)
            data.append(self.row3)

        write(data, self.path)

    def test_init(self):
        """Test that a RawData structure can be initialized"""
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=self.columns)
        self.assertEqual(data.daq_type, self.daq_type)
        self.assertEqual(data.sample_rate, self.sample_rate)
        self.assertEqual(data.columns, self.columns)

    def test_write_with_no_data(self):
        """Test that raw data can be persisted to disk."""
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=self.columns)

        self.assertFalse(self.path.exists())
        write(data, self.path)
        self.assertTrue(self.path.exists())

    def test_load_with_no_data(self):
        """Test that the raw data format can be read by the module."""
        self._write_raw_data()
        data = load(self.path)

        self.assertEqual(data.daq_type, self.daq_type)
        self.assertEqual(data.sample_rate, self.sample_rate)
        self.assertEqual(data.columns, self.columns)
        self.assertEqual(0, len(data.rows))

    def test_load_with_data(self):
        """Test that data can be loaded from a file."""
        self._write_raw_data(include_rows=True)
        loaded_data = load(self.path)

        self.assertEqual(loaded_data.daq_type, self.daq_type)
        self.assertEqual(loaded_data.sample_rate, self.sample_rate)
        self.assertEqual(loaded_data.columns, self.columns)

        self.assertEqual(len(loaded_data.rows), 3)
        self.assertEqual(loaded_data.rows[0], self.row1)
        self.assertEqual(loaded_data.rows[1], self.row2)
        self.assertEqual(loaded_data.rows[2], self.row3)

    def test_load_with_invalid_filepath(self):
        """Test that an exception is raised when an invalid file is loaded."""
        with self.assertRaises(BciPyCoreException):
            path = "invalid/path/to/file.csv"
            load(path)

    def test_deserialization(self):
        """Test that the load function can be accessed through a class
        constructor."""
        self._write_raw_data(include_rows=True)

        loaded_data = RawData.load(self.path)

        self.assertEqual(loaded_data.daq_type, self.daq_type)
        self.assertEqual(loaded_data.sample_rate, self.sample_rate)
        self.assertEqual(loaded_data.columns, self.columns)

        self.assertEqual(len(loaded_data.rows), 3)
        self.assertEqual(loaded_data.rows[0], self.row1)
        self.assertEqual(loaded_data.rows[1], self.row2)
        self.assertEqual(loaded_data.rows[2], self.row3)

    def test_settings(self):
        """Test that metadata settings can be extracted from a raw data
        file."""
        self._write_raw_data()

        name, sample_rate, columns = settings(self.path)
        self.assertEqual(name, self.daq_type)
        self.assertEqual(sample_rate, self.sample_rate)
        self.assertEqual(columns, self.columns)

    def test_raw_data_reader(self):
        """Test that data can be read from a file incrementally."""
        self._write_raw_data(include_rows=True)

        with RawDataReader(self.path) as reader:
            self.assertEqual(reader.daq_type, self.daq_type)
            self.assertEqual(reader.sample_rate, self.sample_rate)
            self.assertEqual(reader.columns, self.columns)

            # all columns are read as strings by default
            self.assertEqual(list(map(str, self.row1)), next(reader))
            self.assertEqual(list(map(str, self.row2)), next(reader))
            self.assertEqual(list(map(str, self.row3)), next(reader))

    def test_raw_data_reader_with_type_conversions(self):
        """Test that data can be read incrementally and that data can be
        converted to numeric types."""
        columns = ['timestamp', 'ch1', 'ch2', 'ch3', 'TRG']
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=columns)

        row1 = [1, 1.0, 2.0, 3.0, 'A']
        row2 = [2, 4.0, 5.0, 6.0, 'B']
        row3 = [3, 7.0, 8.0, 9.0, 'C']

        data.append(row1)
        data.append(row2)
        data.append(row3)
        write(data, self.path)

        with RawDataReader(self.path, convert_data=True) as reader:
            self.assertEqual(reader.daq_type, self.daq_type)
            self.assertEqual(reader.sample_rate, self.sample_rate)
            self.assertEqual(reader.columns, data.columns)

            self.assertEqual(row1, next(reader))
            self.assertEqual(row2, next(reader))
            self.assertEqual(row3, next(reader))

    def test_raw_data_writer(self):
        """Test that data can be written incrementally"""

        rows = [self.row1, self.row2, self.row3]

        self.assertFalse(self.path.exists())
        with RawDataWriter(self.path,
                           daq_type=self.daq_type,
                           sample_rate=self.sample_rate,
                           columns=self.columns) as writer:
            for row in rows:
                writer.writerow(row)
        self.assertTrue(self.path.exists())

    def test_raw_data_writer_writerows(self):
        """Test that data can be written in chunks."""

        rows = [self.row1, self.row2, self.row3]

        self.assertFalse(self.path.exists())
        with RawDataWriter(self.path,
                           daq_type=self.daq_type,
                           sample_rate=self.sample_rate,
                           columns=self.columns) as writer:
            writer.writerows(rows)
        self.assertTrue(self.path.exists())

        loaded_data = RawData.load(self.path)
        self.assertEqual(loaded_data.daq_type, self.daq_type)
        self.assertEqual(loaded_data.sample_rate, self.sample_rate)
        self.assertEqual(loaded_data.columns, self.columns)
        self.assertEqual(len(loaded_data.rows), 3)

    def test_raw_data_writer_initialization(self):
        """Test that writer can be manually opened and closed"""

        rows = [self.row1, self.row2, self.row3]

        self.assertFalse(self.path.exists())
        writer = RawDataWriter(self.path,
                               daq_type=self.daq_type,
                               sample_rate=self.sample_rate,
                               columns=self.columns)
        writer.__enter__()
        writer.writerows(rows)
        writer.__exit__()

        self.assertTrue(self.path.exists())
        loaded_data = RawData.load(self.path)

        self.assertEqual(loaded_data.daq_type, self.daq_type)
        self.assertEqual(loaded_data.sample_rate, self.sample_rate)
        self.assertEqual(loaded_data.columns, self.columns)

        self.assertEqual(len(loaded_data.rows), 3)

    def test_sample_data(self):
        """Test that sample data can be generated for testing purposes."""
        channels = ['ch1', 'ch2', 'ch3']
        data = sample_data(rows=5, ch_names=channels)
        self.assertEqual(5, len(data.rows))
        self.assertEqual(['timestamp', 'ch1', 'ch2', 'ch3', 'TRG'],
                         data.columns)
        self.assertEqual(1, data.rows[0][0])
        self.assertEqual(5, data.rows[4][0])

        # Test data range of channel columns
        for row in data.rows:
            for col in row[1:-1]:
                self.assertTrue(col >= -1000 and col <= 1000)

    def test_sample_data_with_triggers(self):
        """Test generating sample data with triggers at pre-defined timestamps.
        This functionality is used primarily for other unit tests."""
        channels = ['ch1', 'ch2', 'ch3']
        data = sample_data(rows=10,
                           ch_names=channels,
                           triggers=[(1, 'A'), (5, 'B'), (10, 'C')])
        self.assertEqual(10, len(data.rows))

        trg_col = data.columns.index('TRG')
        self.assertEqual('A', data.rows[0][trg_col])
        self.assertEqual('B', data.rows[4][trg_col])
        self.assertEqual('C', data.rows[9][trg_col])

    def test_raw_data_numeric_channels(self):
        """Tests that data channels can be extracted for analysis."""

        columns = ['timestamp', 'ch1', 'ch2', 'ch3', 'TRG']
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=columns)

        row1 = [1, 1.0, 2.0, 3.0, '0.0']
        row2 = [2, 4.0, 5.0, 6.0, 'A']
        row3 = [3, 7.0, 8.0, 9.0, '0.0']

        data.append(row1)
        data.append(row2)
        data.append(row3)

        self.assertEqual(['ch1', 'ch2', 'ch3'], data.channels)

        # Test numeric data frame
        dataframe = pd.DataFrame(data=[[1, 1.0, 2.0, 3.0], [2, 4.0, 5.0, 6.0],
                                       [3, 7.0, 8.0, 9.0]],
                                 columns=['timestamp', 'ch1', 'ch2', 'ch3'])
        self.assertTrue(np.all(dataframe == data.numeric_data))

    def test_data_by_channel(self):
        """Tests that data can be structured in column-order for analysis."""

        columns = ['timestamp', 'ch1', 'ch2', 'TRG']
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=columns)

        data.append([1, 1.0, 2.0, '0.0'])
        data.append([2, 4.0, 5.0, 'A'])
        data.append([3, 7.0, 8.0, '0.0'])

        arr, _ = data.by_channel()
        self.assertEqual((2, 3), arr.shape)

        self.assertTrue(len(data.channels), len(arr))
        self.assertTrue(np.all(arr[0] == [1.0, 4.0, 7.0]),
                        "Should have ch1 data")
        self.assertTrue(np.all(arr[1] == [2.0, 5.0, 8.0]),
                        "Should have ch2 data")

    def test_data_by_channel_applies_transformation(self):
        """Tests that data correctly structured, can have a transformation applied to it."""
        columns = ['timestamp', 'ch1', 'ch2', 'TRG']
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=columns)

        data.append([1, 1.0, 2.0, '0.0'])
        data.append([2, 4.0, 5.0, 'A'])
        data.append([3, 7.0, 8.0, '0.0'])

        transform = mock()
        # note data here should be returned as a nd.array. for mocking we don't care as much
        when(RawData).apply_transform(any(), transform).thenReturn((data, self.sample_rate))
        resp, fs = data.by_channel(transform=transform)

        self.assertEqual(self.sample_rate, fs)
        self.assertEqual(resp, data)
        verify(RawData, times=1).apply_transform(any(), transform)

    def test_data_by_channel_map(self):
        """Tests that when given a channel map, it will filter the numeric columns.

        We assume most trigger columns will be removed (due to being read in as a string or object),
        however some are numeric trigger channels and may require filtering with a channel map.
        Other cases could be dropping bad channels.
        """
        columns = ['timestamp', 'ch1', 'ch2', 'TRG']
        channel_map = [1, 1, 0]
        expected_channels = ['ch1', 'ch2']
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=columns)

        data.append([1, 1.0, 2.0, 0.0])
        data.append([2, 4.0, 5.0, 0.0])
        data.append([3, 7.0, 8.0, 0.0])

        arr, channels, _ = data.by_channel_map(channel_map=channel_map)

        self.assertEqual((2, 3), arr.shape)

        self.assertTrue(len(data.channels), len(arr))
        self.assertTrue(np.all(arr[0] == [1.0, 4.0, 7.0]),
                        "Should have ch1 data")
        self.assertTrue(np.all(arr[1] == [2.0, 5.0, 8.0]),
                        "Should have ch2 data")
        self.assertEqual(channels, expected_channels)

    def test_data_by_channel_map_applies_transformation(self):
        """Tests that when given a channel map, it will filter the numeric columns.

        We assume most trigger columns will be removed (due to being read in as a string or object),
        however some are numeric trigger channels and may require filtering with a channel map.
        Other cases could be dropping bad channels.
        """
        columns = ['timestamp', 'ch1', 'ch2', 'TRG']
        channel_map = [1, 1, 1]
        expected_channels = ['ch1', 'ch2', 'TRG']
        data = RawData(daq_type=self.daq_type,
                       sample_rate=self.sample_rate,
                       columns=columns)

        data.append([1, 1.0, 2.0, 0.0])
        data.append([2, 4.0, 5.0, 0.0])
        data.append([3, 7.0, 8.0, 0.0])

        transform = mock()
        expected_output, expected_fs = data.by_channel()
        # note data here should be returned as a nd.array. for mocking we don't care as much
        when(RawData).by_channel(transform).thenReturn((expected_output, expected_fs))
        _, channels, fs = data.by_channel_map(channel_map=channel_map, transform=transform)

        self.assertEqual(expected_fs, fs)
        self.assertEqual(expected_channels, channels)
        verify(RawData, times=1).by_channel(transform)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import random
from typing import List, Optional
from abc import ABC, abstractmethod

from bcipy.helpers.stimuli import best_selection


class StimuliAgent(ABC):
    @abstractmethod
    def reset(self):
        ...

    @abstractmethod
    def return_stimuli(self, list_distribution: np.ndarray, **kwargs):
        """ updates the agent with most likely posterior and selects queries
            Args:
                list_distribution(list[ndarray]): posterior distributions as
                    stored in the decision maker
            Return:
                query(list[str]): queries """
        ...

    @abstractmethod
    def do_series(self):
        """ If the system decides on a class let the agent know about it """
        ...


class RandomStimuliAgent(StimuliAgent):
    """ An inherited class of StimuliAgent. Chooses random set of letters for
        queries instead of most likely letters.
        Attr:
            alphabet(list[str]): Query space(possible queries).
            len_query(int): number of elements in a query
        Functions:
            reset(): reset the agent
            return_stimuli(): update the agent and return a stimuli set
            do_series(): one a commitment is made update agent
            """

    def __init__(self, alphabet: List[str], len_query: int = 4):
        self.alphabet = alphabet
        self.len_query = len_query

    def reset(self):
        """ This querying method is memoryless no reset needed """
        pass

    def return_stimuli(self, list_distribution: np.ndarray, constants: Optional[List[str]] = None):
        """ return random elements from the alphabet """
        tmp = [i for i in self.alphabet]
        query = random.sample(tmp, self.len_query)

        if constants:
            query[-len(constants):] = constants

        return query

    def do_series(self):
        pass


class NBestStimuliAgent(StimuliAgent):
    """ An inherited class of StimuliAgent. Updates the agent with N most likely
        letters based on posteriors and selects queries.
        Attr:
            alphabet(list[str]): Query space(possible queries).
            len_query(int): number of elements in a query
        Functions:
            reset(): reset the agent
            return_stimuli(): update the agent and return a stimuli set
            do_series(): one a commitment is made update agent
            """

    def __init__(self, alphabet: List[str], len_query: int = 4):
        self.alphabet = alphabet
        self.len_query = len_query

    def reset(self):
        pass

    def return_stimuli(self, list_distribution: np.ndarray, constants: Optional[List[str]] = None):
        p = list_distribution[-1]
        tmp = [i for i in self.alphabet]
        query = best_selection(tmp, p, self.len_query, always_included=constants)

        return query

    def do_series(self):
        pass



"""Tests for list processing utilities"""
import unittest
from bcipy.helpers.list import destutter, grouper


class TestListUtilities(unittest.TestCase):
    """Main test class for list processing"""

    def test_destutter(self):
        """Test removing sequential elements from a list."""
        item1 = dict(a=1, b=1)
        item2 = dict(a=1, b=2)
        item3 = dict(a=2, b=1)
        item4 = dict(a=2, b=2)

        self.assertEqual(destutter([item1, item1, item2, item3, item4]),
                         [item1, item2, item3, item4])
        self.assertEqual(
            destutter([item1, item1, item2, item3, item4],
                      key=lambda x: x['a']), [item2, item4])
        self.assertEqual(
            destutter([item1, item1, item2, item3, item4],
                      key=lambda x: x['b']), [item1, item2, item3, item4])

    def test_grouper_incomplete_fill_default(self):
        iterable = 'ABCDEFG'
        chunk_size = 2
        fillvalue = 'X'
        response = grouper(iterable, chunk_size, fillvalue=fillvalue)
        expected = [('A', 'B'), ('C', 'D'), ('E', 'F'), ('G', 'X')]
        for resp, exp in zip(response, expected):
            self.assertEqual(resp, exp)

    def test_grouper_incomplete_fill_value_error_without_fillvalue(self):
        iterable = 'ABCDEFG'
        chunk_size = 2
        incomplete = 'fill'

        with self.assertRaises(ValueError):
            grouper(iterable, chunk_size, incomplete=incomplete)

    def test_grouper_incomplete_ignore(self):
        iterable = 'ABCDEFG'
        chunk_size = 2
        incomplete = 'ignore'
        expected = [('A', 'B'), ('C', 'D'), ('E', 'F'), ('G')]
        response = grouper(iterable, chunk_size, incomplete=incomplete)
        for resp, exp in zip(response, expected):
            self.assertEqual(resp, exp)

    def test_grouper_incomplete_value_error_unsupported_incompelte_mode_defined(self):
        iterable = 'ABCDEFG'
        chunk_size = 2
        incomplete = 'not_defined'

        with self.assertRaises(ValueError):
            grouper(iterable, chunk_size, incomplete=incomplete)


if __name__ == '__main__':
    unittest.main()



import unittest

from bcipy.config import DEFAULT_EXPERIMENT_ID
from bcipy.helpers.validate import validate_experiment, validate_experiments
from bcipy.helpers.save import save_experiment_data
from bcipy.helpers.exceptions import (
    InvalidExperimentException,
    InvalidFieldException,
    UnregisteredExperimentException,
    UnregisteredFieldException,
)


class TestValidateExperiment(unittest.TestCase):

    def test_validate_experiment_returns_true_on_default(self):
        experiment_name = DEFAULT_EXPERIMENT_ID

        # rely on the default experiment and field path kwargs, pass the experiment_name for validation
        response = validate_experiment(experiment_name)

        self.assertTrue(response)

    def test_validate_experiment_throws_unregistered_expection_on_unregistered_experiment(self):
        experiment_name = 'doesnotexist'
        with self.assertRaises(UnregisteredExperimentException):
            validate_experiment(experiment_name)

    def test_save_experiment_data_throws_unregistered_exception_on_unregistered_fields(self):
        # create a fake experiment to load
        experiment_name = 'test'
        fields = {
            'registered_field': {
                'help_text': 'test',
                'type': 'int'
            }
        }
        experiment = {
            experiment_name: {
                'fields': [{'does_not_exist': {'required': 'false', 'anonymize': 'true'}}], 'summary': ''}
        }

        # assert it raises the expected exception
        with self.assertRaises(UnregisteredFieldException):
            save_experiment_data(experiment, fields, '.', 'test_experiment.json')

    def test_validate_experiment_throws_file_not_found_with_incorrect_experiment_path(self):
        # define an invalid path
        path = 'does/not/exist'

        # assert it raises the expected exception for an invalid experiment_path kwarg
        with self.assertRaises(FileNotFoundError):
            validate_experiment(DEFAULT_EXPERIMENT_ID, experiment_path=path)

    def test_validate_experiment_throws_file_not_found_with_incorrect_field_path(self):
        # define an invalid path
        path = 'does/not/exist'

        # assert it raises the expected exception for an invalid field_path kwarg
        with self.assertRaises(FileNotFoundError):
            validate_experiment(DEFAULT_EXPERIMENT_ID, field_path=path)


class TestValidateExperiments(unittest.TestCase):
    experiment_name = 'test'
    fields = {
        'registered_field': {
            'help_text': 'test',
            'type': 'int'
        }
    }
    experiments = {
        experiment_name: {
            'fields': [{'registered_field': {'required': 'false', 'anonymize': 'true'}}], 'summary': ''}
    }

    def test_validate_experiments_returns_true_on_valid_experiment(self):
        response = validate_experiments(self.experiments, self.fields)

        self.assertTrue(response)

    def test_validate_experiments_throws_invalid_experiment_exception_on_invalid_experiment_no_field(self):
        experiments = {
            'invalid': {
                'summary': ''}
        }
        with self.assertRaises(InvalidExperimentException):
            validate_experiments(experiments, self.fields)

    def test_validate_experiments_throws_invalid_experiment_exception_on_invalid_experiment_invalid_field(self):
        experiments = {
            'invalid': {
                'summary': '',
                'fields': 'should_be_list!'}
        }
        with self.assertRaises(InvalidExperimentException):
            validate_experiments(experiments, self.fields)

    def test_validate_experiments_throws_invalid_experiment_exception_on_invalid_experiment_invalid_summary(self):
        experiments = {
            'invalid': {
                'summary': [],
                'fields': []}
        }
        with self.assertRaises(InvalidExperimentException):
            validate_experiments(experiments, self.fields)

    def test_validate_experiments_throws_invalid_experiment_exception_on_invalid_experiment_no_summary(self):
        experiments = {
            'invalid': {
                'fields': []}
        }
        with self.assertRaises(InvalidExperimentException):
            validate_experiments(experiments, self.fields)

    def test_validate_experiments_throws_invalid_field_exception_on_invalid_field_no_required(self):
        experiments = {
            self.experiment_name: {
                'fields': [{'registered_field': {'anonymize': 'true'}}], 'summary': ''}
        }
        with self.assertRaises(InvalidFieldException):
            validate_experiments(experiments, self.fields)

    def test_validate_experiments_throws_invalid_field_exception_on_invalid_field_no_anonymize(self):
        experiments = {
            self.experiment_name: {
                'fields': [{'registered_field': {'required': 'true'}}], 'summary': ''}
        }
        with self.assertRaises(InvalidFieldException):
            validate_experiments(experiments, self.fields)

    def test_validate_experiments_throws_unregistered_exception_on_unregistered_fields(self):
        experiment = {
            self.experiment_name: {
                'fields': [{'does_not_exist': {'required': 'false', 'anonymize': 'true'}}], 'summary': ''}
        }

        # assert it raises the expected exception
        with self.assertRaises(UnregisteredFieldException):
            validate_experiments(experiment, self.fields)


if __name__ == '__main__':
    unittest.main()



# pylint: disable=fixme
"""Defines the driver for the Device for communicating with
LabStreamingLayer (LSL)."""
import logging
from typing import Dict, List

import pylsl

from bcipy.acquisition.devices import DeviceSpec

log = logging.getLogger(__name__)

LSL_TIMESTAMP = 'LSL_timestamp'
LSL_TIMEOUT_SECONDS = 5.0


class Marker():
    """Data class which wraps a LSL marker; data pulled from a marker stream is
    a tuple where the first item is a list of channels and second item is the
    timestamp. Assumes that marker inlet only has a single channel."""

    def __init__(self, data=(None, None)):
        super(Marker, self).__init__()
        self.channels, self.timestamp = data

    @classmethod
    def empty(cls):
        """Creates an empty Marker."""
        return Marker()

    def __repr__(self):
        return f"<value: {self.trg}, timestamp: {self.timestamp}>"

    @property
    def is_empty(self):
        """Test to see if the current marker is empty."""
        return self.channels is None or self.timestamp is None

    @property
    def trg(self):
        """Get the trigger."""
        # pylint: disable=unsubscriptable-object
        return self.channels[0] if self.channels else None


def inlet_name(inlet) -> str:
    """Returns the name of a pylsl streamInlet."""
    name = '_'.join(inlet.info().name().split())
    return name.replace('-', '_')


def channel_names(stream_info: pylsl.StreamInfo) -> List[str]:
    """Extracts the channel names from the LSL Stream metadata."""
    channels = []
    if stream_info.type() == 'Markers':
        return ['Marker']
    if stream_info.desc().child("channels").empty():
        return channels

    channel = stream_info.desc().child("channels").child("channel")
    for _ in range(stream_info.channel_count()):
        channel_name = channel.child_value("label")
        channels.append(channel_name)
        channel = channel.next_sibling()

    return channels


def check_device(device_spec: DeviceSpec, metadata: pylsl.StreamInfo):
    """Confirm that the properties of the given device_spec match the metadata
    acquired from the device."""
    channels = channel_names(metadata)
    # Confirm that provided channels match metadata, or meta is empty.
    if channels and device_spec.channel_names != channels:
        print(f"device channels: {channels}")
        print(device_spec.channel_names)
        raise Exception("Channels read from the device do not match "
                        "the provided parameters.")
    assert device_spec.channel_count == metadata.channel_count(
    ), "Channel count error"

    if device_spec.sample_rate != metadata.nominal_srate():
        raise Exception("Sample frequency read from device does not match "
                        "the provided parameter")


def rename_items(items: List[str], rules: Dict[str, str]) -> None:
    """Renames items based on the provided rules.
    Parameters
    ----------
        items - list of items ; values will be mutated
        rules - change key -> value
    """
    for key, val in rules.items():
        if key in items:
            items[items.index(key)] = val



"""
This import statement allows users to import submodules from display more directly.

`from bcipy.display import init_display_window` vs. `from bcipy.display.main import init_display_window`
"""
from bcipy.config import BCIPY_LOGO_PATH
from .main import (
    Display,
    InformationProperties,
    init_display_window,
    PreviewInquiryProperties,
    StimuliProperties,
    VEPStimuliProperties
)

__all__ = [
    'Display',
    'init_display_window',
    'BCIPY_LOGO_PATH',
    'StimuliProperties',
    'VEPStimuliProperties',
    'InformationProperties',
    'PreviewInquiryProperties'
]



import logging
import string
from typing import Dict, List, Tuple, Optional

import numpy as np

from bcipy.helpers.stimuli import InquirySchedule, inq_generator, StimuliOrder
from bcipy.helpers.symbols import SPACE_CHAR, BACKSPACE_CHAR
from bcipy.task.control.query import RandomStimuliAgent, StimuliAgent
from bcipy.task.control.criteria import CriteriaEvaluator

log = logging.getLogger(__name__)


class EvidenceFusion():
    """ Fuses likelihood evidences provided by the inference
        Attr:
            evidence_history(dict{list[ndarray]}): Dictionary of difference
                evidence types in list. Lists are ordered using the arrival
                time.
            likelihood(ndarray[]): current probability distribution over the
                set. Gets updated once new evidence arrives. """

    def __init__(self, list_name_evidence, len_dist):
        self.evidence_history = {name: [] for name in list_name_evidence}
        self.likelihood = np.ones(len_dist) / len_dist

    def update_and_fuse(self, dict_evidence):
        """ Updates the probability distribution
            Args:
                dict_evidence(dict{name: ndarray[float]}): dictionary of
                    evidences (EEG and other likelihoods)
        """

        for key in dict_evidence.keys():
            tmp = dict_evidence[key][:][:]
            self.evidence_history[key].append(tmp)

        # Current rule is to multiply
        for value in dict_evidence.values():
            self.likelihood *= value[:]

        if np.isinf(np.sum(self.likelihood)):
            tmp = np.zeros(len(self.likelihood))
            tmp[np.where(self.likelihood == np.inf)[0][0]] = 1
            self.likelihood = tmp

        if not np.isnan(np.sum(self.likelihood)):
            self.likelihood = self.likelihood / np.sum(self.likelihood)

        likelihood = self.likelihood[:]

        return likelihood

    def reset_history(self):
        """ Clears evidence history """
        for value in self.evidence_history.values():
            del value[:]
        self.likelihood = np.ones(len(self.likelihood)) / len(self.likelihood)

    def save_history(self) -> None:
        """ Saves the current likelihood history """
        log.warning('save_history not implemented')
        return

    @property
    def latest_evidence(self) -> Dict[str, List[float]]:
        """Latest evidence of each type in the evidence history.

        Returns
        -------
        a dictionary with an entry for all configured evidence types.
        """
        return {
            name: list(evidence[-1]) if evidence else []
            for name, evidence in self.evidence_history.items()
        }


class DecisionMaker:
    """ Scheduler of the entire framework
        Attr:
            state(str): state of the framework, which increases in size
                by 1 after each inquiry. Elements are alphabet, ".,_,<"
                where ".": null_inquiry(no decision made)
                      "_": space bar
                      "<": back space
            alphabet(list[str]): list of symbols used by the framework. Can
                be switched with location of images or one hot encoded images.
            is_txt_stim(bool): whether the stimuli are text or images
            inq_constants(list[str]): optional list of letters which should appear in
                every inquiry.
            stopping_evaluator: CriteriaEvaluator - optional parameter to
                provide alternative rules for committing to a decision.
            stimuli_agent(StimuliAgent): the query selection mechanism of the
                system
            stimuli_timing(list[float]): list of timings for the stimuli ([fixation_time, stimuli_flash_time])
            stimuli_order(StimuliOrder): ordering of the stimuli (random, distributed)
            stimuli_jitter(float): jitter of the inquiry stimuli in seconds

        Functions:
            decide():
                Checks the criteria for making and series, using all
                evidences and decides to do an series or to collect more
                evidence
            do_series():
                Once committed an series perform updates to condition the
                distribution on the previous letter.
            schedule_inquiry():
                schedule the next inquiry using the current information
            decide_state_update():
                If committed to an series update the state using a decision
                metric.
                (e.g. pick the letter with highest likelihood)
            prepare_stimuli():
                prepares the query set for the next inquiry
                (e.g pick n-highest likely letters and randomly shuffle)
        """

    def __init__(self,
                 state: str = '',
                 alphabet: List[str] = list(string.ascii_uppercase) + [BACKSPACE_CHAR] + [SPACE_CHAR],
                 is_txt_stim: bool = True,
                 stimuli_timing: List[float] = [1, .2],
                 stimuli_jitter: float = 0,
                 stimuli_order: StimuliOrder = StimuliOrder.RANDOM,
                 inq_constants: Optional[List[str]] = None,
                 stopping_evaluator: CriteriaEvaluator = CriteriaEvaluator.default(min_num_inq=2,
                                                                                   max_num_inq=10,
                                                                                   threshold=0.8),
                 stimuli_agent: Optional[StimuliAgent] = None):
        self.state = state
        self.displayed_state = self.form_display_state(state)
        self.stimuli_timing = stimuli_timing
        self.stimuli_order = stimuli_order
        self.stimuli_jitter = stimuli_jitter

        self.alphabet = alphabet
        self.is_txt_stim = is_txt_stim

        self.list_series = [{'target': None, 'time_spent': 0,
                             'list_sti': [], 'list_distribution': [],
                             'decision': None}]
        self.time = 0
        self.inquiry_counter = 0

        self.stopping_evaluator = stopping_evaluator
        self.stimuli_agent = stimuli_agent or RandomStimuliAgent(alphabet=self.alphabet)
        self.last_selection = ''

        # Items shown in every inquiry
        self.inq_constants = inq_constants

    def reset(self, state=''):
        """ Resets the decision maker with the initial state
            Args:
                state(str): current state of the system """
        self.state = state
        self.displayed_state = self.form_display_state(self.state)

        self.list_series = [{'target': None, 'time_spent': 0,
                             'list_sti': [], 'list_distribution': []}]
        self.time = 0
        self.inquiry_counter = 0

        self.stimuli_agent.reset()

    def form_display_state(self, state):
        """ Forms the state information or the user that fits to the
            display. Basically takes '.' and BACKSPACE_CHAR into consideration and rewrites
            the state
            Args:
                state(str): state string
            Return:
                displayed_state(str): state without '<,.' and removes
                    backspaced letters """
        tmp = ''
        for i in state:
            if i == BACKSPACE_CHAR:
                tmp = tmp[0:-1]
            elif i != '.':
                tmp += i

        return tmp

    def update(self, state=''):
        self.state = state
        self.displayed_state = self.form_display_state(state)

    def decide(self, p) -> Tuple[bool, InquirySchedule]:
        """ Once evidence is collected, decision_maker makes a decision to
            stop or not by leveraging the information of the stopping
            criteria. Can decide to do an series or schedule another inquiry.

        Args
        ----
        p(ndarray[float]): |A| x 1 distribution array
            |A|: cardinality of the alphabet

        Return
        ------
        - commitment: True if a letter is a commitment is made
        False if requires more evidence
        - inquiry schedule: Extra arguments depending on the decision
        """

        self.list_series[-1]['list_distribution'].append(p[:])

        # Check stopping criteria
        if self.stopping_evaluator.should_commit(self.list_series[-1]):
            self.do_series()
            return True, None
        else:
            stimuli = self.schedule_inquiry()
            return False, stimuli

    def do_series(self):
        """ series refers to a commitment to a decision.
            If made, state is updated, displayed state is updated
            a new series is appended. """
        self.inquiry_counter = 0
        decision = self.decide_state_update()
        self.last_selection = decision
        self.state += decision
        self.displayed_state = self.form_display_state(self.state)

        # Initialize next series
        self.list_series.append({'target': None, 'time_spent': 0,
                                 'list_sti': [], 'list_distribution': []})

        self.stimuli_agent.do_series()
        self.stopping_evaluator.do_series()

    def schedule_inquiry(self) -> InquirySchedule:
        """ Schedules next inquiry """
        self.state += '.'
        stimuli = self.prepare_stimuli()
        self.list_series[-1]['list_sti'].append(stimuli[0])
        self.inquiry_counter += 1

        return stimuli

    def decide_state_update(self):
        """ Checks stopping criteria to commit to an series """
        idx = np.where(
            self.list_series[-1]['list_distribution'][-1] ==
            np.max(self.list_series[-1]['list_distribution'][-1]))[0][0]
        decision = self.alphabet[idx]
        self.list_series[-1]['decision'] = decision
        return decision

    def prepare_stimuli(self) -> InquirySchedule:
        """ Given the alphabet, under a rule, prepares a stimuli for
            the next inquiry.

        Return
        ------
        stimuli(tuple[list[char],list[float],list[str]]): tuple of
        stimuli information. [0]: letter, [1]: timing, [2]: color
        """

        # querying agent decides on possible letters to be shown on the screen
        query_els = self.stimuli_agent.return_stimuli(
            self.list_series[-1]['list_distribution'],
            constants=self.inq_constants)
        # once querying is determined, append with timing and color info.
        stimuli = inq_generator(query=query_els,
                                inquiry_count=1,
                                is_txt=self.is_txt_stim,
                                timing=self.stimuli_timing,
                                stim_order=self.stimuli_order,
                                stim_jitter=self.stimuli_jitter)
        return stimuli



from typing import Optional, Tuple

import numpy as np
from scipy.signal import butter, filtfilt, iirnotch, sosfiltfilt


class Notch:
    """Remove a single frequency"""

    def __init__(self, sample_rate_hz, remove_freq_hz=60.0, quality_factor=30):
        remove_freq_hz = remove_freq_hz / (sample_rate_hz / 2)
        self.b, self.a = iirnotch(remove_freq_hz, quality_factor)

    def __call__(self, data: np.ndarray, fs: Optional[int] = None) -> Tuple[np.ndarray, int]:
        return filtfilt(self.b, self.a, data), fs


class Bandpass:
    """Preserve a specified range of frequencies"""

    def __init__(self, lo, hi, sample_rate_hz, order=5):
        nyq = 0.5 * sample_rate_hz
        lo, hi = lo / nyq, hi / nyq
        self.sos = butter(order, [lo, hi], analog=False, btype="band", output="sos")

    def __call__(self, data: np.ndarray, fs: Optional[int] = None) -> Tuple[np.ndarray, int]:
        return sosfiltfilt(self.sos, data), fs


def filter_inquiries(inquiries, transform, sample_rate) -> Tuple[np.ndarray, float]:
    """Filter Inquiries.

    The shape of data after reshaping into inquiries requires a bit of pre-processing to apply the
        transforms and filters defined in BciPy without looping. Here we flatten the inquires, filter,
        and return them in the correct shape for continued processing.
    """
    # (Channels, Inquiries, Samples)
    old_shape = inquiries.shape
    # (Channels*Inquiry, Samples)
    inq_flatten = inquiries.reshape(-1, old_shape[-1])
    inq_flatten_filtered, transformed_sample_rate = transform(inq_flatten, sample_rate)
    # (Channels, Inquiries, Samples)
    inquiries = inq_flatten_filtered.reshape(*old_shape[:2], inq_flatten_filtered.shape[-1])
    return inquiries, transformed_sample_rate



import logging
from typing import List, NamedTuple, Optional, Tuple

from psychopy import core

from bcipy.config import (
    SESSION_DATA_FILENAME, TRIGGER_FILENAME, WAIT_SCREEN_MESSAGE, SESSION_SUMMARY_FILENAME)
from bcipy.display import (InformationProperties, PreviewInquiryProperties,
                           StimuliProperties)
from bcipy.display import Display
from bcipy.display.paradigm.rsvp.mode.copy_phrase import CopyPhraseDisplay
from bcipy.display.components.task_bar import CopyPhraseTaskBar
from bcipy.feedback.visual.visual_feedback import VisualFeedback
from bcipy.helpers.clock import Clock
from bcipy.helpers.copy_phrase_wrapper import CopyPhraseWrapper
from bcipy.helpers.exceptions import TaskConfigurationException
from bcipy.helpers.list import destutter
from bcipy.helpers.save import _save_session_related_data
from bcipy.helpers.session import session_excel
from bcipy.helpers.stimuli import InquirySchedule, StimuliOrder
from bcipy.helpers.task import (construct_triggers,
                                fake_copy_phrase_decision,
                                get_data_for_decision, get_user_input,
                                target_info, trial_complete_message)
from bcipy.helpers.symbols import BACKSPACE_CHAR, alphabet
from bcipy.helpers.triggers import (FlushFrequency, Trigger, TriggerHandler,
                                    TriggerType, convert_timing_triggers)
from bcipy.signal.model.inquiry_preview import compute_probs_after_preview
from bcipy.task import Task
from bcipy.task.data import EvidenceType, Inquiry, Session


class Decision(NamedTuple):
    """Represents the result of evaluating evidence.

    Attrs
    -----
    - decision_made : whether or a not there was a commitment to a letter.
    - selection : selected letter; will be an empty string if there was no
    commitment.
    - spelled_text : spelled text resulting from the decision.
    - new_inq_schedule : the next inquiry to present if there was not a
    decision.
    """
    decision_made: bool
    selection: str
    spelled_text: str
    new_inq_schedule: Optional[InquirySchedule]


class RSVPCopyPhraseTask(Task):
    """RSVP Copy Phrase Task.

    Initializes and runs all needed code for executing a copy phrase task. A
        phrase is set in parameters and necessary objects (daq, display) are
        passed to this function. Certain Wrappers and Task Specific objects are
        executed here.

    Parameters
    ----------
        win : object,
            display window to present visual stimuli.
        daq : object,
            data acquisition object initialized for the desired protocol
        parameters : dict,
            configuration details regarding the experiment. See parameters.json
        file_save : str,
            path location of where to save data from the session
        signal_model : loaded pickle file,
            trained signal model.
        language_model: object,
            trained language model.
        fake : boolean, optional
            boolean to indicate whether this is a fake session or not.
    Returns
    -------
        file_save : str,
            path location of where to save data from the session
    """

    TASK_NAME = 'RSVP Copy Phrase Task'
    MODE = 'RSVP'

    PARAMETERS_USED = [
        'time_fixation', 'time_flash', 'time_prompt', 'trial_length',
        'font', 'fixation_color', 'trigger_type',
        'filter_high', 'filter_low', 'filter_order', 'notch_filter_frequency', 'down_sampling_rate', 'prestim_length',
        'is_txt_stim', 'lm_backspace_prob', 'backspace_always_shown',
        'decision_threshold', 'max_inq_len', 'max_inq_per_series', 'max_minutes', 'max_selections', 'min_inq_len',
        'show_feedback', 'feedback_duration',
        'show_preview_inquiry', 'preview_inquiry_isi',
        'preview_inquiry_key_input', 'preview_inquiry_length', 'preview_inquiry_progress_method',
        'spelled_letters_count', 'static_trigger_offset',
        'stim_color', 'stim_height', 'stim_jitter', 'stim_length', 'stim_number',
        'stim_order', 'stim_pos_x', 'stim_pos_y', 'stim_space_char', 'target_color',
        'task_buffer_length', 'task_color', 'task_height', 'task_text',
        'info_pos_x', 'info_pos_y', 'info_color', 'info_height', 'info_text', 'info_color', 'info_height', 'info_text',
    ]

    def __init__(self, win, daq, parameters, file_save, signal_model,
                 language_model, fake):
        super(RSVPCopyPhraseTask, self).__init__()

        self.window = win
        self.daq = daq
        self.parameters = parameters

        self.validate_parameters()

        self.static_clock = core.StaticPeriod(
            screenHz=self.window.getActualFrameRate())
        self.experiment_clock = Clock()
        self.start_time = self.experiment_clock.getTime()

        self.alp = alphabet(self.parameters)

        self.button_press_error_prob = 0.05
        self.evidence_types = [EvidenceType.LM, EvidenceType.ERP]
        if self.parameters['show_preview_inquiry']:
            self.evidence_types.append(EvidenceType.BTN)

        self.file_save = file_save

        self.trigger_handler = TriggerHandler(self.file_save, TRIGGER_FILENAME, FlushFrequency.EVERY)
        self.session_save_location = f"{self.file_save}/{SESSION_DATA_FILENAME}"
        self.copy_phrase = parameters['task_text']

        self.fake = fake
        self.language_model = language_model
        self.signal_model = signal_model
        self.evidence_precision = 5

        self.feedback = VisualFeedback(
            self.window,
            {'feedback_font': self.parameters['font'],
             'feedback_color': self.parameters['info_color'],
             'feedback_pos_x': self.parameters['info_pos_x'],
             'feedback_pos_y': self.parameters['info_pos_y'],
             'feedback_stim_height': self.parameters['info_height'],
             'feedback_duration': self.parameters['feedback_duration']},
            self.experiment_clock)

        self.setup()

        # set a preview_only parameter
        self.parameters.add_entry(
            'preview_only',
            {
                'value': 'true' if self.parameters['preview_inquiry_progress_method'] == 0 else 'false',
                'section': '',
                'readableName': '',
                'helpTip': '',
                'recommended_values': '',
                'type': 'bool'
            }
        )

        self.rsvp = self.init_display()

    def setup(self) -> None:
        """Initialize/reset parameters used in the execute run loop."""

        self.spelled_text = str(
            self.copy_phrase[0:self.starting_spelled_letters()])
        self.last_selection = ''
        self.inq_counter = 0
        self.session = Session(
            save_location=self.file_save,
            task='Copy Phrase',
            mode=self.MODE,
            symbol_set=self.alp,
            decision_threshold=self.parameters['decision_threshold'])
        self.write_session_data()

        self.init_copy_phrase_task()
        self.current_inquiry = self.next_inquiry()

    def init_display(self) -> Display:
        """Initialize the display"""
        return _init_copy_phrase_display(self.parameters, self.window,
                                         self.static_clock,
                                         self.experiment_clock,
                                         self.spelled_text)

    def validate_parameters(self) -> None:
        """Validate.

        Confirm Task is configured with correct parameters and within operating limits.
        """

        # ensure all required parameters are provided
        for param in RSVPCopyPhraseTask.PARAMETERS_USED:
            if param not in self.parameters:
                raise TaskConfigurationException(f"parameter '{param}' is required")

        # ensure data / query parameters are set correctly
        buffer_len = self.parameters['task_buffer_length']
        prestim = self.parameters['prestim_length']
        poststim = self.parameters['trial_length']
        if buffer_len < prestim:
            raise TaskConfigurationException(
                f'task_buffer_length=[{buffer_len}] must be greater than prestim_length=[{prestim}]')

        if buffer_len < poststim:
            raise TaskConfigurationException(
                f'task_buffer_length=[{buffer_len}] must be greater than trial_length=[{poststim}]')

    def starting_spelled_letters(self) -> int:
        """Number of letters already spelled at the start of the task."""
        spelled_letters_count = self.parameters['spelled_letters_count']
        if spelled_letters_count > len(self.copy_phrase):
            self.logger.debug('Already spelled letters exceeds phrase length.')
            spelled_letters_count = 0
        return spelled_letters_count

    def next_inquiry(self) -> Optional[InquirySchedule]:
        """Generate an InquirySchedule for spelling the next letter."""
        if self.copy_phrase_task:
            _, inquiry_schedule = self.copy_phrase_task.initialize_series()
            return inquiry_schedule
        return None

    def init_copy_phrase_task(self) -> None:
        """Initialize the CopyPhraseWrapper.

        Returns:
        --------
        initialized CopyPhraseWrapper
        """

        self.copy_phrase_task = CopyPhraseWrapper(
            self.parameters['min_inq_len'],
            self.parameters['max_inq_per_series'],
            lmodel=self.language_model,
            device_spec=self.daq.device_spec,
            signal_model=self.signal_model,
            k=self.parameters['down_sampling_rate'],
            alp=self.alp,
            evidence_names=self.evidence_types,
            task_list=[(str(self.copy_phrase), self.spelled_text)],
            is_txt_stim=self.parameters['is_txt_stim'],
            stim_timing=[
                self.parameters['time_fixation'], self.parameters['time_flash']
            ],
            decision_threshold=self.parameters['decision_threshold'],
            backspace_prob=self.parameters['lm_backspace_prob'],
            backspace_always_shown=self.parameters['backspace_always_shown'],
            filter_high=self.parameters['filter_high'],
            filter_low=self.parameters['filter_low'],
            filter_order=self.parameters['filter_order'],
            notch_filter_frequency=self.parameters['notch_filter_frequency'],
            stim_length=self.parameters['stim_length'],
            stim_jitter=self.parameters['stim_jitter'],
            stim_order=StimuliOrder(self.parameters['stim_order']))

    def user_wants_to_continue(self) -> bool:
        """Check if user wants to continue or terminate.

        Returns
        -------
        - `True` to continue
        - `False` to finish the task.
        """
        should_continue = get_user_input(
            self.rsvp,
            WAIT_SCREEN_MESSAGE,
            self.parameters['stim_color'],
            first_run=self.first_run)
        if not should_continue:
            self.logger.debug('User wants to exit.')
        return should_continue

    def wait(self, seconds: float = None):
        """Pause for a time.

        Parameters
        ----------
        - seconds : duration of time to wait; if missing, defaults to the
        value of the parameter `'task_buffer_length'`
        """
        seconds = seconds or self.parameters['task_buffer_length']
        core.wait(seconds)

    def present_inquiry(self, inquiry_schedule: InquirySchedule
                        ) -> Tuple[List[Tuple[str, float]], bool]:
        """Present the given inquiry and return the trigger timing info.

        Parameters
        ----------
        - inquiry_schedule : schedule for next sequences of stimuli to present.
        Currently, only the first list of stims in the schedule is used.

        Returns
        -------
        Tuple of `(stim_times, proceed)`

        - stim_times : list of tuples representing the stimulus and time that
        it was presented relative to the experiment clock. Non-stim triggers
        may be also be included in the list ('calibration', etc).
        - proceed : indicates whether to proceed with evaluating eeg evidence.
        a value of False indicates that the inquiry was previewed but not
        presented in sequence.
        """
        # Update task state and reset the static
        self.rsvp.update_task_bar(self.spelled_text)
        self.rsvp.draw_static()
        self.window.flip()

        self.wait()

        # Setup the new stimuli
        self.rsvp.schedule_to(stimuli=inquiry_schedule.stimuli[0],
                              timing=inquiry_schedule.durations[0],
                              colors=inquiry_schedule.colors[0]
                              if self.parameters['is_txt_stim'] else None)

        if self.parameters['show_preview_inquiry']:
            stim_times, proceed = self.rsvp.preview_inquiry()
            if proceed:
                stim_times.extend(self.rsvp.do_inquiry())
        else:
            stim_times = self.rsvp.do_inquiry()
            proceed = True

        return stim_times, proceed

    def show_feedback(self, selection: str, correct: bool = True):
        """Display the selection as feedback if the 'show_feedback'
        parameter is configured.

        Parameters
        ----------
        - selection : selected stimulus to display
        - correct : whether or not the correct stim was chosen
        """
        if self.parameters['show_feedback']:
            self.feedback.administer(f'Selected: {selection}')

    def check_stop_criteria(self) -> bool:
        """Returns True if experiment is currently within params and the task
        should continue.
        """
        if self.copy_phrase == self.spelled_text:
            self.logger.debug('Spelling complete')
            return False

        if (self.inq_counter + 1) >= self.parameters['max_inq_len']:
            self.logger.debug('Max tries exceeded: to allow for more tries'
                              ' adjust the Maximum inquiry Length '
                              '(max_inq_len) parameter.')
            return False

        if self.session.total_time_spent >= (self.parameters['max_minutes'] *
                                             60):
            self.logger.debug('Max time exceeded. To allow for more time '
                              'adjust the max_minutes parameter.')
            return False

        if self.session.total_number_decisions >= self.parameters['max_selections']:
            self.logger.debug('Max number of selections reached '
                              '(configured with the max_selections parameter)')
            return False

        return True

    def next_target(self) -> str:
        """Computes the next target letter based on the currently spelled_text.
        """
        if self.copy_phrase[0:len(self.spelled_text)] == self.spelled_text:
            # if correctly spelled so far, get the next letter.
            return self.copy_phrase[len(self.spelled_text)]
        return BACKSPACE_CHAR

    def execute(self) -> str:
        """Executes the task.

        Returns
        -------
        data save location (triggers.txt, session.json)
        """
        self.logger.debug('Starting Copy Phrase Task!')
        run = True
        self.wait()  # buffer for data processing

        while run and self.user_wants_to_continue() and self.current_inquiry:
            target_stimuli = self.next_target()
            stim_times, proceed = self.present_inquiry(
                self.current_inquiry)

            self.write_trigger_data(stim_times, target_stimuli)
            self.wait()

            evidence_types = self.add_evidence(stim_times, proceed)
            decision = self.evaluate_evidence()

            data = self.new_data_record(stim_times,
                                        target_stimuli,
                                        current_text=self.spelled_text,
                                        decision=decision,
                                        evidence_types=evidence_types)
            self.update_session_data(data,
                                     save=True,
                                     decision_made=decision.decision_made)

            if decision.decision_made:
                self.show_feedback(decision.selection,
                                   (decision.selection == target_stimuli))
                self.spelled_text = decision.spelled_text
                self.current_inquiry = self.next_inquiry()

            else:
                self.current_inquiry = decision.new_inq_schedule

            run = self.check_stop_criteria()
            self.inq_counter += 1

        self.exit_display()
        self.write_offset_trigger()

        self.session.task_summary = TaskSummary(
            self.session, self.parameters['show_preview_inquiry'],
            self.parameters['preview_inquiry_progress_method'],
            self.trigger_handler.file_path).as_dict()
        self.write_session_data()

        # Evidence is not recorded in the session when using fake decisions.
        if self.parameters['summarize_session'] and self.session.has_evidence():
            session_excel(session=self.session,
                          excel_file=f"{self.file_save}/{SESSION_SUMMARY_FILENAME}")

        # Wait some time before exiting so there is trailing eeg data saved
        self.wait()

        return self.file_save

    def evaluate_evidence(self) -> Decision:
        """Uses the `copy_phrase_task` parameter to evaluate the provided
        evidence and attempt a decision.

        Modifies
        --------
        - self.copy_phrase_task
        """
        if self.fake:
            _, spelled, _ = fake_copy_phrase_decision(self.copy_phrase,
                                                      self.next_target(),
                                                      self.spelled_text)
            # Reset the stoppage criteria by forcing the commit to a decision.
            self.copy_phrase_task.decision_maker.do_series()
            # In fake mode, only the LM is providing evidence, so the decision
            # made is the highest symbol predicted. Override this state
            self.copy_phrase_task.decision_maker.update(spelled)

            # In fake mode, all inquiries result in a selection.
            return Decision(decision_made=True,
                            selection=spelled[-1],
                            spelled_text=spelled,
                            new_inq_schedule=None)

        decision_made, new_sti = self.copy_phrase_task.decide()
        spelled_text = self.copy_phrase_task.decision_maker.displayed_state
        selection = ''
        if decision_made:
            selection = self.copy_phrase_task.decision_maker.last_selection

        return Decision(decision_made, selection, spelled_text, new_sti)

    def add_evidence(self, stim_times: List[List],
                     proceed: bool = True) -> List[EvidenceType]:
        """Add all evidence used to make a decision.

        Parameters
        ----------
        - stim_times : list of stimuli returned from the display
        - proceed : whether or not to proceed with the inquiry

        Returns
        -------
        list of evidence types added

        Modifies
        --------
        - self.copy_phrase_task
        """
        evidences = [
            self.compute_button_press_evidence(proceed),
            self.compute_eeg_evidence(stim_times, proceed)
        ]
        evidence_types = []
        for evidence in evidences:
            if evidence:
                evidence_type, probs = evidence
                evidence_types.append(evidence_type)
                self.copy_phrase_task.add_evidence(evidence_type, probs)
        if self.session.latest_series_is_empty():
            evidence_types.append(EvidenceType.LM)
        return evidence_types

    def compute_button_press_evidence(
            self, proceed: bool) -> Optional[Tuple[EvidenceType, List[float]]]:
        """If 'show_preview_inquiry' feature is enabled, compute the button
        press evidence and add it to the copy phrase task.

        Parameters
        ----------
            proceed : whether to proceed with the inquiry after the preview

        Returns
        -------
            tuple of (evidence type, evidence) or None if inquiry preview is
            not enabled.
        """
        if not self.parameters['show_preview_inquiry'] \
                or not self.current_inquiry \
                or self.parameters['preview_only']:
            return None
        probs = compute_probs_after_preview(self.current_inquiry.stimuli[0],
                                            self.alp,
                                            self.button_press_error_prob,
                                            proceed)
        return (EvidenceType.BTN, probs)

    def compute_eeg_evidence(self,
                             stim_times: List[List],
                             proceed: bool = True
                             ) -> Optional[Tuple[EvidenceType, List[float]]]:
        """Evaluate the EEG evidence and add it to the copy_phrase_task, but
        don't yet attempt a decision.

        Parameters
        ----------
        - stim_times : list of stimuli returned from the display
        - proceed : whether or not to evaluate the evidence, if `False` returns
        empty values.

        Returns
        -------
        tuple of (evidence type, evidence)
        """
        if not proceed or self.fake:
            return None

        # currently prestim_length is used as a buffer for filter application, use the same at the end of the inquiry
        post_stim_buffer = self.parameters['prestim_length']
        raw_data, triggers = get_data_for_decision(
            inquiry_timing=self.stims_for_decision(stim_times),
            daq=self.daq,
            offset=self.parameters['static_trigger_offset'],
            prestim=self.parameters['prestim_length'],
            poststim=post_stim_buffer + self.parameters['trial_length'])

        # we assume all are nontargets at this point
        labels = ['nontarget'] * len(triggers)

        probs = self.copy_phrase_task.evaluate_eeg_evidence(
            raw_data, triggers, labels, self.parameters['trial_length'])
        return (EvidenceType.ERP, probs)

    def stims_for_decision(self, stim_times: List[List]) -> List[List]:
        """The stim_timings from the display may include non-letter stimuli
        such as calibration and inquiry_preview timings. This method extracts
        only the letter data used to process the data for a decision.

        Parameters
        ----------
        - stim_times : list of [stim, clock_time] pairs returned from display.

        Returns
        -------
        stim times where the stim is in the current alphabet; filters out
        'calibration', 'inquiry_preview', etc.
        """
        return [
            timing for timing in stim_times if timing[0] in (self.alp + ['+'])
        ]

    def new_data_record(self,
                        stim_times: List[List],
                        target_stimuli: str,
                        current_text: str,
                        decision: Decision,
                        evidence_types: List[EvidenceType] = None) -> Inquiry:
        """Construct a new inquiry data record.

        Parameters
        ----------
        - stim_times : list of [stim, clock_time] pairs returned from display.
        - target_stimuli : stim the user is currently attempting to spell.
        - current_text : spelled text before the inquiry
        - decision : decision made by the decision maker
        - evidence_types : evidence provided to the decision-maker during the
        current inquiry.

        Returns
        -------
        Inquiry data for the current schedule; returned value only contains
        evidence for the provided evidence_types, leaving the other types empty
        """
        assert self.current_inquiry, "Current inquiry is required"
        evidence_types = evidence_types or []
        triggers = construct_triggers(self.stims_for_decision(stim_times))
        data = Inquiry(stimuli=self.current_inquiry.stimuli,
                       timing=self.current_inquiry.durations,
                       triggers=triggers,
                       target_info=target_info(triggers, target_stimuli,
                                               self.parameters['is_txt_stim']),
                       target_letter=target_stimuli,
                       current_text=current_text,
                       target_text=self.copy_phrase,
                       selection=decision.selection,
                       next_display_state=decision.spelled_text)
        data.precision = self.evidence_precision

        if not self.fake:
            latest_evidence = self.copy_phrase_task.conjugator.latest_evidence
            data.evidences = {
                ev_type: evidence if ev_type in evidence_types else []
                for ev_type, evidence in latest_evidence.items()
            }
            data.likelihood = list(self.copy_phrase_task.conjugator.likelihood)
        return data

    def exit_display(self):
        """Close the UI and cleanup."""
        # Update task state and reset the static
        self.rsvp.update_task_bar(text=self.spelled_text)

        # Say Goodbye!
        self.rsvp.info_text = trial_complete_message(self.window, self.parameters)
        self.rsvp.draw_static()
        self.window.flip()

        # Give the system time to process
        self.wait()

    def update_session_data(self,
                            data: Inquiry,
                            save: bool = True,
                            decision_made: bool = False) -> None:
        """Update the current session with the latest inquiry data

        Parameters
        ----------
        - data : inquiry to append to the session
        - save : if True, persists the session to disk.

        Modifies
        --------
        - self.session
        """
        self.session.add_sequence(data)
        self.session.total_time_spent = self.experiment_clock.getTime() - self.start_time
        if save:
            self.write_session_data()
        if decision_made:
            self.session.add_series()

    def write_session_data(self) -> None:
        """Save session data to disk."""
        if self.session:
            session_file = _save_session_related_data(
                self.session_save_location,
                self.session.as_dict())
            session_file.close()

    def write_offset_trigger(self) -> None:
        """Append the offset to the end of the triggers file.
        """
        if self.daq.is_calibrated:
            self.trigger_handler.add_triggers(
                [Trigger(
                    'daq_sample_offset',
                    TriggerType.SYSTEM,
                    # to help support future refactoring or use of lsl timestamps only
                    # we write only the sample offset here
                    self.daq.offset(self.rsvp.first_stim_time)
                )])
        self.trigger_handler.close()

    def write_trigger_data(self, stim_times: List[Tuple[str, float]], target_stimuli: str) -> None:
        """Save trigger data to disk.

        Parameters
        ----------
        - stim_times : list of (stim, clock_time) tuples
        - target_stimuli : current target stimuli
        """
        if self.first_run and self.daq.is_calibrated:
            # write offset first
            self.trigger_handler.add_triggers(
                [Trigger(
                    'starting_offset',
                    TriggerType.OFFSET,
                    # offset will factor in true offset and time relative from beginning
                    (self.daq.offset(self.rsvp.first_stim_time) - self.rsvp.first_stim_time)
                )]
            )

        triggers = convert_timing_triggers(stim_times, target_stimuli, self.trigger_type)
        self.trigger_handler.add_triggers(triggers)

    def trigger_type(self, symbol: str, target: str, index: int) -> TriggerType:
        """Trigger Type.

        Cast a given symbol to a TriggerType.
        """
        if symbol == 'inquiry_preview':
            return TriggerType.PREVIEW
        if 'bcipy_key_press' in symbol:
            return TriggerType.EVENT
        if symbol == '+':
            return TriggerType.FIXATION
        if target == symbol:
            return TriggerType.TARGET
        return TriggerType.NONTARGET

    def name(self) -> str:
        return self.TASK_NAME

    @property
    def first_run(self) -> bool:
        """First run.

        Determines whether it is the first inquiry presentation / run.
        """
        return self.inq_counter == 0


class TaskSummary:
    """Summary data for tracking performance metrics.

    Parameters
    ----------
        session - current session information
        show_preview - whether or not inquiry preview was displayed
        preview_mode - the switch mode for inquiry preview:
            0 = preview only;
            1 = press to confirm;
            2 = press to skip to another inquiry
    """

    def __init__(self,
                 session: Session,
                 show_preview: bool = False,
                 preview_mode: int = 0,
                 trigger_path: str = None):
        assert preview_mode in range(3), 'Preview mode out of range'
        self.session = session
        self.show_preview = show_preview
        self.preview_mode = preview_mode
        self.trigger_path = trigger_path
        self.logger = logging.getLogger(__name__)

    def as_dict(self) -> dict:
        """Computes the task summary data to append to the session."""

        selections = [
            inq for inq in self.session.all_inquiries if inq.selection
        ]
        correct = [inq for inq in selections if inq.is_correct_decision]
        incorrect = [inq for inq in selections if not inq.is_correct_decision]

        # Note that SPACE is considered a symbol
        correct_symbols = [
            inq for inq in correct if inq.selection != BACKSPACE_CHAR
        ]

        btn_presses = self.btn_press_count()
        sel_count = len(selections)
        switch_per_selection = (btn_presses /
                                sel_count) if sel_count > 0 else 0
        accuracy = (len(correct) / sel_count) if sel_count > 0 else 0

        # Note that minutes includes startup time and any breaks.
        minutes = self.session.total_time_spent / 60
        return {
            'selections_correct': len(correct),
            'selections_incorrect': len(incorrect),
            'selections_correct_symbols': len(correct_symbols),
            'switch_total': btn_presses,
            'switch_per_selection': switch_per_selection,
            'switch_response_time': self.switch_response_time(),
            'typing_accuracy': accuracy,
            'correct_rate': len(correct) / minutes if minutes else 0,
            'copy_rate': len(correct_symbols) / minutes if minutes else 0
        }

    def btn_press_count(self) -> int:
        """Compute the number of times the switch was activated. Returns 0 if
        inquiry preview mode was off or mode was preview-only."""

        if not self.show_preview or self.preview_mode == 0:
            return 0

        inquiries = self.session.all_inquiries
        if self.preview_mode == 1:
            # press to confirm
            activations = [inq for inq in inquiries if inq.eeg_evidence]
        elif self.preview_mode == 2:
            # press to skip
            activations = [inq for inq in inquiries if not inq.eeg_evidence]
        return len(activations)

    def switch_response_time(self) -> Optional[float]:
        """Computes the average switch response in seconds."""

        # Remove consecutive items with the same type; we are only interested
        # in PREVIEW followed by a EVENT.
        triggers = destutter(self.switch_triggers(), key=lambda trg: trg.type)
        pairs = list(zip(triggers[::2], triggers[1::2]))

        # Confirm that the data is structured as expected.
        for preview, keypress in pairs:
            if (preview.type != TriggerType.PREVIEW) or (
                    keypress.type != TriggerType.EVENT):
                self.logger.info('Could not compute switch_response_time')
                return None

        response_times = [
            keypress.time - preview.time for preview, keypress in pairs
        ]
        count = len(response_times)
        return sum(response_times) / count if count > 0 else None

    def switch_triggers(self) -> List[Trigger]:
        """Returns a list of switch-related triggers"""
        if not self.trigger_path:
            return []
        triggers, _offset = TriggerHandler.read_text_file(self.trigger_path)
        return [
            trg for trg in triggers
            if trg.type in [TriggerType.PREVIEW, TriggerType.EVENT]
        ]


def _init_copy_phrase_display(parameters, win, static_clock, experiment_clock, starting_spelled_text):
    preview_inquiry = PreviewInquiryProperties(
        preview_only=parameters['preview_only'],
        preview_inquiry_length=parameters['preview_inquiry_length'],
        preview_inquiry_key_input=parameters['preview_inquiry_key_input'],
        preview_inquiry_progress_method=parameters[
            'preview_inquiry_progress_method'],
        preview_inquiry_isi=parameters['preview_inquiry_isi'])
    info = InformationProperties(
        info_color=[parameters['info_color']],
        info_pos=[(parameters['info_pos_x'], parameters['info_pos_y'])],
        info_height=[parameters['info_height']],
        info_font=[parameters['font']],
        info_text=[parameters['info_text']],
    )
    stimuli = StimuliProperties(stim_font=parameters['font'],
                                stim_pos=(parameters['stim_pos_x'],
                                          parameters['stim_pos_y']),
                                stim_height=parameters['stim_height'],
                                stim_inquiry=['A'] * parameters['stim_length'],
                                stim_colors=[parameters['stim_color']] * parameters['stim_length'],
                                stim_timing=[10] * parameters['stim_length'],
                                is_txt_stim=parameters['is_txt_stim'])

    task_bar = CopyPhraseTaskBar(win,
                                 task_text=parameters['task_text'],
                                 spelled_text=starting_spelled_text,
                                 colors=[parameters['task_color']],
                                 font=parameters['font'],
                                 height=parameters['task_height'])

    return CopyPhraseDisplay(win,
                             static_clock,
                             experiment_clock,
                             stimuli,
                             task_bar,
                             info,
                             starting_spelled_text,
                             trigger_type=parameters['trigger_type'],
                             space_char=parameters['stim_space_char'],
                             preview_inquiry=preview_inquiry)



"""Utilities for system information and general functionality that may be
shared across modules."""
import importlib
import logging
import os
import pkgutil
import platform
import socket
import sys
import time
import torch
from pathlib import Path
from typing import Callable, List, Optional, NamedTuple

import pkg_resources
import psutil
import pyglet
from cpuinfo import get_cpu_info

from bcipy.config import DEFAULT_ENCODING, LOG_FILENAME


class ScreenInfo(NamedTuple):
    width: int
    height: int
    rate: float


def is_connected(hostname: str = "1.1.1.1", port=80) -> bool:
    """Test for internet connectivity.

    Parameters
    ----------
        hostname - name of host for attempted connection
        port - port on host to which to connect
    """
    try:
        conn = socket.create_connection(address=(hostname, port), timeout=2)
        conn.close()
        return True
    except OSError:
        pass
    return False


def is_battery_powered() -> bool:
    """Check if this current computer is a laptop currently using its battery.

    Returns
    -------
    True if the computer is currently running on battery power. This can impact
    the performance of hardware (ex. GPU) needed for BciPy operation by entering
    power saving operations."""
    return psutil.sensors_battery(
    ) and not psutil.sensors_battery().power_plugged


def is_screen_refresh_rate_low(refresh: Optional[float] = None) -> bool:
    """Check if the screen refresh rate is sufficient for BciPy operation.

    Parameters
    ----------
        refresh(float) - optional screen refresh rate if already collected. If not provided,
            the current screen refresh rate is gathered using get_screen_info().

    Returns
    -------
    True if the refresh rate is less than 120 Hz; otherwise False.
    """
    if refresh is None:
        refresh = get_screen_info().rate

    return refresh < 120


def git_dir() -> str:
    """Git Directory.

    Returns the root directory with the .git folder. If this source code
    was not checked out from scm, answers None."""

    # Relative to current file; may need to be modified if method is moved.
    git_root = Path(os.path.abspath(__file__)).parent.parent.parent
    git_meta = Path(os.path.join(git_root, '.git'))
    return os.path.abspath(git_meta) if git_meta.is_dir() else None


def git_hash() -> Optional[str]:
    """Git Hash.

    Returns an abbreviated git sha hash if this code is being run from a
    git cloned version of bcipy; otherwise returns an empty string.

    Could also consider making a system call to:
    git describe --tags
    """

    git_path = git_dir()
    if not git_path:
        print('.git path not found')
        return None

    try:
        head_path = Path(os.path.join(git_path, 'HEAD'))
        with open(head_path, encoding=DEFAULT_ENCODING) as head_file:
            # First line contains a reference to the current branch.
            # ex. ref: refs/heads/branch_name
            ref = head_file.readline()

        ref_val = ref.split(':')[-1].strip()
        ref_path = Path(os.path.join(git_path, ref_val))
        with open(ref_path, encoding=DEFAULT_ENCODING) as ref_file:
            sha = ref_file.readline()

        # sha hash is 40 characters; use an abbreviated 7-char version which
        # is displayed in github.
        return sha[0:7]
    except Exception as error:
        print(f'Error reading git version: {error}')
        return None


def bcipy_version() -> str:
    """BciPy Version.

    Gets the current bcipy version. If the current instance of bcipy is a
    git repository, appends the current abbreviated sha hash.
    """
    version = pkg_resources.get_distribution('bcipy').version
    sha_hash = git_hash()

    return f'{version} - {sha_hash}' if sha_hash else version


def get_screen_info() -> ScreenInfo:
    """Gets the screen information.

    Note: Use this method if only the screen resolution is needed; it is much more efficient
    than extracting that information from the dict returned by the get_system_info method.

    Returns
    -------
        ScreenInfo(width, height, rate)
     """
    screen = pyglet.canvas.get_display().get_default_screen()
    return ScreenInfo(screen.width, screen.height, screen.get_mode().rate)


def get_gpu_info() -> List[dict]:
    """Information about GPUs available for processing."""
    properties = []
    for idx in range(torch.cuda.device_count()):
        prop = torch.get_device_properties(idx)
        properties.append(dict(name=prop.name, total_memory=prop.total_memory))
    return properties


def get_system_info() -> dict:
    """Get System Information.
    See: https://stackoverflow.com/questions/3103178/how-to-get-the-system-info-with-python

    Returns
    -------
        dict of system-related properties, including ['os', 'py_version', 'resolution',
          'memory', 'bcipy_version', 'platform', 'platform-release', 'platform-version',
          'architecture', 'processor', 'cpu_count', 'hz', 'ram']
    """
    screen_info = get_screen_info()
    info = get_cpu_info()
    gpu_info = get_gpu_info()
    return {
        'os': sys.platform,
        'py_version': sys.version,
        'screen_resolution': [screen_info.width, screen_info.height],
        'memory': psutil.virtual_memory().available / 1024 / 1024,
        'bcipy_version': bcipy_version(),
        'platform': platform.system(),
        'platform-release': platform.release(),
        'platform-version': platform.version(),
        'architecture': platform.machine(),
        'processor': platform.processor(),
        'cpu_count': os.cpu_count(),
        'cpu_brand': info['brand_raw'],
        'cpu_hz': info['hz_actual_friendly'],
        'gpu_available': torch.cuda.is_available(),
        'gpu_count': len(gpu_info),
        'gpu_details': gpu_info,
        'screen_refresh': f"{screen_info.rate}Hz",
        'ram': str(round(psutil.virtual_memory().total / (1024.0**3))) + " GB"
    }


def configure_logger(
        save_folder: str,
        log_name=LOG_FILENAME,
        log_level=logging.DEBUG,
        version=None) -> None:
    """Configure Logger.

    Does what it says.
    """
    # create the log file
    logfile = os.path.join(save_folder, 'logs', log_name)

    # configure it
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    handler = logging.FileHandler(logfile, 'w', encoding='utf-8')
    handler.setFormatter(logging.Formatter(
        '[%(threadName)-9s][%(asctime)s][%(name)s][%(levelname)s]: %(message)s'))
    root_logger.addHandler(handler)

    print(f'Printing all BciPy logs to: {logfile}')

    if version:
        logging.info(f'Start of Session for BciPy Version: ({version})')


def import_submodules(package, recursive=True):
    """Import Submodules.

    Import all submodules of a module, recursively, including subpackages.
    https://stackoverflow.com/questions/3365740/how-to-import-all-submodules

    Parameters
    ----------
        package : str | package
            name of package or package instance
        recursive : bool, optional

    Returns
    -------
        dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}
    for _loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        if name.startswith('test'):
            continue
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results


def dot(relative_to: str, *argv) -> str:
    """
    Given a file location and one or more subdirectory/filename args, returns
    a Path as a str for that file.

    Ex. dot(__file__, 'fst', 'brown_closure.n5.kn.fst')
    """
    working_dir = Path(os.path.dirname(os.path.realpath(relative_to)))
    for subdir in argv:
        # uses the '/' operator in pathlib to construct a new Path.
        working_dir = working_dir / subdir
    return str(working_dir)


def auto_str(cls):
    """Autogenerate a str method to print all variable names and values.
    https://stackoverflow.com/questions/32910096/is-there-a-way-to-auto-generate-a-str-implementation-in-python
    """

    def __str__(self):
        return '%s(%s)' % (type(self).__name__, ', '.join(
            '%s=%s' % item for item in vars(self).items()))

    cls.__str__ = __str__
    return cls


def log_to_stdout():
    """Set logging to stdout. Useful for demo scripts.
    https://stackoverflow.com/questions/14058453/making-python-loggers-output-all-messages-to-stdout-in-addition-to-log-file
    """

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '[%(threadName)-9s][%(asctime)s][%(name)s][%(levelname)s]: %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)


def report_execution_time(func: Callable) -> Callable:
    """Report execution time.

    A decorator to log execution time of methods in seconds. To use,
        decorate your method with @report_execution_time.
    """
    log = logging.getLogger()

    def wrap(*args, **kwargs):
        time1 = time.perf_counter()
        response = func(*args, **kwargs)
        time2 = time.perf_counter()
        log.info('{:s} method took {:0.4f}s to execute'.format(func.__name__, (time2 - time1)))
        return response
    return wrap
